// $Id: networker_main.C,v 1.1.1.1 2007/04/09 19:05:53 pjungkun Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL
 
#include "Global.H"
#include "Game.H"
#include "GUI.H"
#include "CycleCounter.H"
#include "Options.H"
#include "Server.H"
#include "GameStateModule.H"
#include "Compress.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "MapTool.H"
#include "Net_Worker.H"

using namespace std;


const char *ORTS_VERSION = "ORTS version 3.1 (2005-May-16)";

#define USE_EFENCE 0

extern int EF_PROTECT_FREE, EF_FILL, EF_PROTECT_BELOW, EF_ALLOW_MALLOC_0;
 
static const sint4 GFX_X = 900;  // graphics window size
static const sint4 GFX_Y = 740;

static void add_options()
{
  Options o("Networker");

  o.put("-loadmap", string(""),      "map file name to load");
  o.put("-newt",    true,            "new terrain");
  o.put("-debug",   false,           "debug mode");

  o.put("-disp",    "2d display");

  o.put("-m",       string("server"),"mode: server|client|game|test|worker");
  o.put("-showv",   false,           "show views");
  o.put("-showl",   false,           "show message length");

  o.put("-hfile",   string(""),      "create map from height/ramp file");
  o.put("-out",     string(""),      "save map to file");
  
  o.put("-username",string("worker"),             "login name for the worker");
  o.put("-password",string("password"),           "password for the worker");
  o.put("-server",  string("opal.cs.ualberta.ca"),"GGS server IP address");
  o.put("-sport",   sint4(5000),                  "GGS server port");
  o.put("-servicename", string("service"),                "name of the netservice");
}


void create_orts_world(stringstream &world)
{
  // note: if loading a map, -nplayers values must match
  sint4 num_players;
  Options::get("-nplayers", num_players);
  
  // note: current design is for the map to contain the blueprint,
  // thus ignoring any -bp value on the orts command-line.
  // If there is no map specified, the -bp value is used.
  string map_fname;
  Options::get("-loadmap", map_fname);
  
  if (map_fname.empty()) {

    string hfile;
    Options::get("-hfile", hfile);

    if (!hfile.empty()) {

      // read height map + ramp information from file
      
      MapTool::read_height_map(world);

    } else {

      // default behaviour - create random world using seed value

      bool newt, debug;
      Options::get("-newt", newt);
      Options::get("-debug", debug);
      
      if (debug)
        MapTool::generate_debug_map(world);
      else if (newt)
	MapTool::generate_random_cliff_map(world);
      else 
	MapTool::generate_random_map(world);
    }
   
  } else {

    // load world from map file
    cout << "loading world from map file: " << map_fname << endl;

    ifstream map_file(map_fname.c_str());
    if (!map_file.is_open()) ERR ("file not found, cannot load map");

    while (map_file.good()) {
      char ch;
      map_file.get(ch);
      world << ch;
    }
  
    map_file.close();
  }
}


struct ClientEventHandler : public EventHandler
{
  GameStateModule *gsm;
  GUI *gui;
  
  ClientEventHandler(GameStateModule *gsm_, GUI *gui_) : gsm(gsm_), gui(gui_) { }
  
  bool handle_event(const Event &e)
  {
    if (e.get_who() == GameStateModule::FROM) {

      if (e.get_what() == GameStateModule::STOP_MSG) exit(0);
      
      if (e.get_what() == GameStateModule::VIEW_MSG) {

	cout << "." << flush;
	
	if (gui) {
	  gui->event();
	  gui->display();
	  if (gui->quit) exit(0);
	}

	gsm->send_actions();
	
	return true;

      } else errstr << "unhandled event: from "
		    << e.get_who() <<  " : "
		    << e.get_what() << endl;
    }

    return false;
  }

};


struct ServerEventHandler : public EventHandler
{
  Server *server;
  GUI *gui;
  bool init;
  
  ServerEventHandler(Server *server_, GUI *gui_) :
    server(server_), gui(gui_), init(false) {}
  
  bool handle_event(const Event& e)
  {
    if (gui && !init) {
      init = true;
      gui->init(GFX_X, GFX_Y, server->get_game());
    }
    
    if (e.get_who() == Server::FROM) {

      if (e.get_what() == Server::STEP_MSG) {

	if (gui) {
	  gui->event();
	  gui->display();
	  if (gui->quit) exit(0);
	}
       
	return true;

      } else
	
	errstr << "unhandled event: from "
	       << e.get_who() <<  " : " << e.get_what()
	       << endl;
    }
    return false;
  }

};


int main(int argc, char **argv)
{
#if USE_EFENCE
  // aggressive efence checks
  EF_PROTECT_FREE = 1;
  EF_FILL = 0xee;
  EF_PROTECT_BELOW = 1;
  EF_ALLOW_MALLOC_0 = 0;
#endif

  signals_shut_down(true);

  add_options();
  MapTool::Options::add();
  GameStateModule::Options::add();
  Server::Options::add();
  GUI::add_options();

  if (Options::process(argc, argv, cerr, ORTS_VERSION)) exit(20);

  sint4 seed;
  Options::get("-seed", seed);
  if (!seed) seed = time(0);
  Options::set("-seed", seed);
  cout << "seed=" << seed << endl;
  
  GUI *gui = 0;
  bool disp;
  Options::get("-disp", disp);

  if (disp) {
    SDLinit::video_init();
    gui = new GUI;
  }
 
  string mode;
  Options::get("-m", mode);
  
  if (mode == "server" || mode == "client")
    SDLinit::network_init();

  if (mode == "client") {

    GameStateModule::Options gsmo;
    GameStateModule gsm(gsmo);
    ClientEventHandler ceh(&gsm, gui);

    gsm.add_handler(&ceh);
    
    if (!gsm.connect()) exit(10);

    if (gui) {
      gui->init(GFX_X, GFX_Y, gsm.get_game());
      gui->display();
    }

    while (!shut_down) {
      if (!gsm.recv_view()) SDL_Delay(1);
    }
    
    return 0;
  }

  stringstream world;
  create_orts_world(world);

  // up to here copied from orts_main.C
  
  Net_Worker worker(world.str());
  worker.run();
    
  cout << "game finished" << endl;
  
  return 0;
}



#if 0

const char *ORTS_VERSION = "ORTS version 1.9 (2004-Oct-5)";

#define USE_EFENCE 0

extern int EF_PROTECT_FREE, EF_FILL, EF_PROTECT_BELOW, EF_ALLOW_MALLOC_0;

#define GFX_X 900  // graphics window size
#define GFX_Y 740

static void add_options(Options &o)
{
  o.put("-bp",      string("stdgame/units.bp"), "blueprints file to load");
  o.put("-loadmap", string(""),      "map file name to load");
  o.put("-water",                    "create water");
  o.put("-newt",    true,            "new terrain");
  o.put("-debug",   false,           "debug mode");

  o.put("-disp",    "display game");
  o.put("-dvisi",   "mark visible tiles");
  o.put("-dknow",   "mark known tiles");
  o.put("-dseen",   "mark which players can see each object");
  o.put("-dgrid",   "show the tile grid");

  o.put("-m",       string("server"),"mode: server|client|game|test|worker");
  o.put("-showv",   false,           "show views");
  o.put("-showl",   false,           "show message length");

  o.put("-x",       sint4(56),       "playing field width");
  o.put("-y",       sint4(38),       "playing field height");
  o.put("-fplat",   real8(0.6),      "fraction of terrain that is plateau");
  o.put("-clump",   sint4(800),      "degree to which terrain is clustered (1=none)");
  o.put("-smooth",  true,            "smooth borders");
  o.put("-holes",   true,            "try harder to plug terrain holes");
  o.put("-hfile",   string(""),      "create map from height/ramp file");
  o.put("-out",     string(""),      "save map to file");
  o.put("-savereplay", string(""),   "save simulation to replay file (server only)");

  o.put("-nobjs",   sint4(100),      "number of objects per player");
  o.put("-nplayers", sint4(1),       "number of players");
}


struct ClientEventHandler : public EventHandler
{
  Client *client;
  GUI *gui;
  
  ClientEventHandler(Client *client_, GUI *gui_);
  
  sint4 handle_event(const Event& e)
  {
    if (e.who() == Client::FROM_CLIENT) {

      if (e.what() == Client::STOP_MSG) exit(0);
      
      if (e.what() == Client::VIEW_MSG) {
	if (gui) {
	  gui->event();
	  gui->display();
	  if (gui->quit) exit(0);
	}

	return 1;

      } else errstr << "unhandled event: from "
		    << e.who() <<  " : "
		    << e.what() << endl;
    }

    return 0;
  }

};


struct ServerEventHandler : public EventHandler
{
  Server *server;
  GUI *gui;
  bool init;
  
  ServerEventHandler(Server *server_, GUI *gui_) :
    server(server_), gui(gui_), init(false) {}
  
  sint4 handle_event(const Event& e)
  {
    if (gui && !init) {
      init = true;
      gui->init(GFX_X, GFX_Y, server->get_game());
    }
    
    if (e.who() == Server::FROM_SERVER) {

      if (e.what() == Server::STEP_MSG) {

	if (gui) {
	  gui->event();
	  gui->display();
	  if (gui->quit) exit(0);
	}
       
	return 1;

      } else
	
	errstr << "unhandled event: from "
	       << e.who() <<  " : " << e.what()
	       << endl;
    }
    return 0;
  }

};


int main(int argc, char **argv)
{

#if USE_EFENCE
  // aggressive efence checks
  EF_PROTECT_FREE = 1;
  EF_FILL = 0xee;
  EF_PROTECT_BELOW = 1;
  EF_ALLOW_MALLOC_0 = 0;
#endif

  signals_shut_down(true);

  Options opt;
  add_options(opt);
  kernel_add_options(opt);
  serverclient_add_options(opt);

  if (Options::process(argc, argv, cerr, ORTS_VERSION)) exit(20);

  kernel_init_rnd(opt);
  
  SDLinit::network_init();

  stringstream world;
  create_orts_world(opt, world);

  Net_Worker worker(world.str(), opt);
  worker.run();
    
  cout << "game finished" << endl;
  
  return 0;
}

#endif
