#include "GameStateModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "Game.H"
#include "GameObj.H"
#include "SimpleTerrain.H"
#include "TerrainModule.H"

/**
 * This program demos path finding and execution using the TerrainModule.
 * It randomly moves the first marine it finds.  If there is no marine, it does nothing.
 *
 * TerrainModule Events
 *   User Requests
 *     FIND_PATH_MSG: request to move to a location
 *     FIND_PATH_STOP: request to abort move
 *   Info returned by TerrainModule
 *     PATH_SUCCESS: goal reached
 *     PATH_FAILURE: TerrainModule giving up on goal due to...
 *       PATH_PLAN_FAILURE: A* being unable to find a path (returned in details field)
 *       PATH_MOVE_FAILURE: Inability to follow the path found by A* (returned in details field)
 *
 *  Here's how SimpleTerrain.C responds to a unit's failure to execute a move action:
 *    Unexpected stop. The current policy is:
 *      First time => wait a little and send the move command again.
 *      Second stop at the same place => mark as failed and replan.
 *      Third stop at the same place => report failure.
 *
 *   Interesting supporting classes for TerrainModule: 
 *     TerrainBase, TerrainBasicImp, SimpleTerrain, ST_PFEngine 
 *
 */

using namespace std;

bool quit = false;
GameStateModule *gsmP;

// Our event handler is also a ClientModule so it can emit events,
//   specifically sending movement requests to TerrainModule
class SampleEventHandler : public EventHandler, public ClientModule {
public:
	const sint4 FROM;
	SampleEventHandler ()  : FROM( EventFactory::new_who() ), marine(0) {}
	bool handle_event(const Event &e) { 
		if (e.get_who() == GameStateModule::FROM
			&& e.get_what() == GameStateModule::STOP_MSG)
			quit = true;
		if (e.get_who() == GameStateModule::FROM
			&& e.get_what() == GameStateModule::VIEW_MSG) {
			if (!marine)
				getMarine();
		}
		if ( e.get_who() == TerrainModule::FROM ) {
			const TerrainModule::PathStatusEvent* pse = 
				dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
			if( pse ) {
				GameObj* gob = pse->get_obj();
				assert (gob == marine); // gob useful when planning paths for multiple units
				sint4 type = pse->get_what();
				if (type == TerrainModule::PATH_SUCCESS)
					cout << "TerrainModule::PATH_SUCCESS" << endl;
				else if (type == TerrainModule::PATH_FAILURE) {
					cout << "TerrainModule::PATH_FAILURE" << endl;
					sint4 why = pse->get_details();
					if (why == TerrainModule::PATH_PLAN_FAILURE)
						cout << "TerrainModule::PATH_PLAN_FAILURE" << endl;
					else if (why == TerrainModule::PATH_MOVE_FAILURE)
						cout << "TerrainModule::PATH_MOVE_FAILURE" << endl;
					else {
						cout << "Path failure for unknown reason" << endl;
						assert(0); // Should never get here
					}
				}
				else {
					cout << "Message from TerrainModule of unknown type" << endl;
					assert(0); // Should never get here
				}
				// No matter what happened...let's just request another move!
				requestMove();
			}
			return true;
		}
	}
private:
	const GameObj *marine;

	void getMarine()
	{
		Game &game = gsmP->get_game();
		int playerID = game.get_client_player();
		const Game::ObjCont &objs = game.get_objs(playerID);
		FORALL (objs, it) {
			const GameObj *gob = (*it)->get_GameObj();
			if ( gob->bp_name() == "marine" ) {
				cout << "Found marine at " << *gob->sod.x << " " << *gob->sod.y  << endl;
				marine = gob;
				break;
			}
		}
		requestMove();
	}

	void requestMove()
	{
		Game &game = gsmP->get_game();
		const Map<GameTile> &map = game.get_map();
		const sint4 points_x = map.get_width()  * game.get_tile_points();
		const sint4 points_y = map.get_height() * game.get_tile_points(); 
		cout << "Playing field dimensions: " << points_x << " " << points_y << endl;
		sint4 target_x = rand() % points_x;
		sint4 target_y = rand() % points_y;
		TerrainBase::Loc location(target_x,target_y);
		call_handlers( TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG,
			TerrainBase::Obj2LocTask( (Object*)marine, location)) );
		cout << "Requesting move to " << target_x << " " << target_y << endl;
	}
};


int main(int argc, char **argv)
{
	signals_shut_down(true);  
	SDLinit::network_init();

	// Create option objects (though we pass them no data) and create modules
	GameStateModule::Options::add();
	Game::Options::add();  
	SimpleTerrain::ST_Terrain::add_options();
	GameStateModule::Options gsmo;
	GameStateModule gsm(gsmo);
	gsmP = &gsm;

	SimpleTerrain::ST_Terrain timp;
	TerrainModule tm(gsm, timp, 110);

	SampleEventHandler* psev = new SampleEventHandler();

	// gsm events are handled in sev
	gsm.add_handler(psev);

	// Terrain module and sev exchange events
	psev->add_handler( &tm );
	tm.add_handler( psev );

	// Connect game state module to server
	if (!gsm.connect()) ERR("connection problems");

	while ( ! quit )
		if (!gsm.recv_view()) SDL_Delay(1);
		else gsm.send_actions(); // Required to send at least one action message

	return 0;
}
