// $Id: Game1_ST_PFEngine.C,v 1.1.1.1 2007/04/09 19:06:50 pjungkun Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "Global.H"
#include "Options.H"
#include "Game1_ST_PFEngine.H"
#include "Distance.H"

using namespace std;
using boost::array;
using boost::shared_ptr;

#define TEST(x)
//#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x

namespace Game1_SimpleTerrain
{
  //-----------------------------------------------------------------------------
  static Loc get_center_loc(const Object *obj)
  {
    Loc l;
    obj->get_center(l.x, l.y);
    return l;
  }

  //-----------------------------------------------------------------------------
  /* Due to our internal low resolution, a location which is clear might appear
     occupied after the conversion to inner coords. So we must accept as "arrived" 
     locations that are a little distant from the exact goal. By how much?
      * Obstacle centers might appear translated and closer to us by ME
        (max_error) in each axis => sqrt(2) * ME
      * Our radius might appear larger by 1 * ME.
      * For circular obstacles, their radius might also appear larger by 1 * ME.
      * Since we consider each int coord as a full cell, two adjacent cells
        are actually (resolution_factor=2*ME) away from each other in world 
        coords.
  */
  class ST_Terrain::PFEngine::IsNearLocation {
    const Loc goal;
    const sint4 allow_distance;
  public:
    IsNearLocation(const Loc &goal_, sint4 allow_distance_=0) 
      : goal(goal_), allow_distance(allow_distance_) {};

    bool operator()(sint4 x, sint4 y, const Map &map) const {
      return Loc(map.x2world(x),map.y2world(y)).distance(goal) <= 
                (5.4142 * map.max_error_from_conversion() + allow_distance) ;
    };

    bool test_for_world_coords(sint4 worldx, sint4 worldy) const {
      return goal.distance(Loc(worldx, worldy)) <= allow_distance;
    };
  };

  //-----------------------------------------------------------------------------
  /* Due to our internal low resolution, adjacent objects in the world coords 
      might appear a little distant after conversion to inner coords and back.
      * Let the moving object be exactly at xy2world(x,y) (since we are 
        sending it there).         
      * The target might appear translated by (max_error) in each axis.
        => sqrt(2) * ME
      * Both target's and obj's radii might appear larger, each by 
        at most (max_error) => + 2*ME
      * If the target is a rectangle, the coords of each side are converted
        seperately, so only the translation effect is relevant for it, no 
        radius effect.
      * Since we consider each int coord as a full cell, two adjacent cells
        are actually (resolution_factor=2*ME) away from each other in world coords.
      * Consider "touch" as distance <= 1 in world coords.
  */
  class ST_Terrain::PFEngine::IsNearObject {
    mutable ObjectTranslocator moving;
    const Object *goal;
    const real8 allowed_ME_units;
    const sint4 allow_distance;
  public:
    IsNearObject(const Object *moving_, const Object *goal_, sint4 allow_distance_=0)
      : moving(moving_), goal(goal_),
        // see above: sqrt(2) + 2 + 1 [+ 1]
        allowed_ME_units(4.4142 + ((goal_->get_shape() == Object::CIRCLE)? 1:0)),
        allow_distance(allow_distance_) {};
    IsNearObject(const IsNearObject& other)
      : moving(other.moving), goal(other.goal), 
        allowed_ME_units(other.allowed_ME_units), allow_distance(other.allow_distance) {};

    bool operator()(sint4 x, sint4 y, const Map &map) const {
      moving.set_position(map.x2world(x), map.y2world(y));
      return Distance::distance(*goal, moving) <= 
        ((allowed_ME_units * map.max_error_from_conversion()) + 1 + allow_distance);
    };

    bool test_for_world_coords(sint4 worldx, sint4 worldy) const {
      moving.set_position(worldx, worldy);
      // Defining "touch" as distance <= 1:
      return Distance::distance(*goal, moving) <= (1 + allow_distance);
    };
  private:
    IsNearObject& operator=(const IsNearObject&);
  };

  //-----------------------------------------------------------------------------
  // When PFing on one (high resolution) map, try to get to a location that is 
  // unoccupied in another (low resolution) map.
  class ST_Terrain::PFEngine::IsAtClearLocation {
    const shared_ptr<const Map> test_clear_on_map;
    mutable Bucket obj_bucket;
  public:
    IsAtClearLocation(const shared_ptr<const Map>& test_clear_on_map_, const Object *obj) 
      : test_clear_on_map(test_clear_on_map_), 
        obj_bucket(BucketFactory::get_bucket(test_clear_on_map_->world2r(obj->get_radius()))) {};
    IsAtClearLocation(const IsAtClearLocation &other)
      : test_clear_on_map(other.test_clear_on_map), 
        obj_bucket(other.obj_bucket) {};

    bool operator()(sint4 x, sint4 y, const Map &PF_map) const {
      // translate from the PF map to world coords
      return test_for_world_coords(PF_map.x2world(x), PF_map.y2world(y));
    };

    bool test_for_world_coords(sint4 worldx, sint4 worldy) const {
      uint4 stayCost = 0;
      // translate from world coords to the test map coords
      sint4 x = test_clear_on_map->world2x(worldx);
      sint4 y = test_clear_on_map->world2y(worldy);
      obj_bucket.set_center(Loc(x,y));
      obj_bucket.check_current_location(*test_clear_on_map, stayCost);
      return stayCost < (uint4)UNIT_COST && stayCost < (uint4)BOUND_COST;
    }
  private:
    IsAtClearLocation& operator=(const IsAtClearLocation&);
  };

  //-----------------------------------------------------------------------------
  ST_Terrain::PFEngine::PFEngine(sint4 tiles_x_, sint4 tiles_y_, 
                                 sint4 tile_points_, const Vector<sint4>& lods_needed)
    : lods(lods_needed)
  {
    assert(lods.size() >= 1 && "Must supply at least one level-of-detail");
    assert(*min_element(lods.begin(), lods.end()) > 0 && "All levels-of-detail must be positive numbers");

    // We need a full resolution version for our internal use (see "clearing" in find_path())
    if( *min_element(lods.begin(), lods.end()) > 1 )
      lods.push_back(1);

    // Remember the most detailed level for use in refining the final leg of paths
    // when approaching a target, and for clearing away from adjacent objects.
    max_details_LOD = min_element(lods.begin(), lods.end()) - lods.begin();

    FORALL(lods, level_of_detail) {
      // Verify parameters
      assert( (tiles_x_ * tile_points_ % (*level_of_detail) == 0) && "Width must be divisible by all levels of detail");
      assert( (tiles_y_ * tile_points_ % (*level_of_detail) == 0) && "Height must be divisible by all levels of detail");

      // Create (and init if necessary) the maps for this LOD:
      sint4 w = tiles_x_ * tile_points_;
      sint4 h = tiles_y_ * tile_points_;
      shared_ptr<Map> ground( new Map(w, h, *level_of_detail) );
      init_sizes(*ground.get());
      ground_maps.push_back(ground);
      
      shared_ptr<Map> air( new Map(w, h, *level_of_detail) );
      init_sizes(*air.get());
      air_maps.push_back(air);
      
      shared_ptr<AStarMap> astar( new AStarMap(w, h, *level_of_detail) );
      astar_maps.push_back(astar);
    }
  }
/*    : map(tiles_x_ * gran + 1, tiles_y_ * gran + 1),
      air_map(tiles_x_ * gran + 1, tiles_y_ * gran + 1),
      astar_map(tiles_x_ * gran + 1, tiles_y_ * gran + 1),
      world_tile_points(tile_points_),
      wtp_per_cell(tile_points_/gran),
      disp(wtp_per_cell/2)
  {
    init_sizes(map);
    init_sizes(air_map);    
  }
*/
  //-----------------------------------------------------------------------------
  bool ST_Terrain::PFEngine::find_path(const Object *obj,
                                       const Loc &goal_loc,
                                       Path& output,
                                       sint4 distance, 
                                       LevelOfDetail lod,
                                       ConsiderObjects consider,
				       uint4 stop_time)
  {
    TEST( cout << "Planning for " << obj << " from " << get_center_loc(obj) << " to " << goal_loc << endl;);
    return sub_path(obj, Loc(0,0), Loc(-1,-1), goal_loc, output, distance, lod, consider, stop_time);
  }

  //-----------------------------------------------------------------------------
  bool ST_Terrain::PFEngine::find_path(const Object *obj,
                                       const Object *goal_obj,
                                       Path& output,
                                       sint4 distance,
                                       LevelOfDetail lod,
                                       ConsiderObjects consider,
				       uint4 stop_time)
  {
    TEST( cout << "Planning for " << obj << " from " << get_center_loc(obj) << " to " << goal_obj << " @ " << get_center_loc(goal_obj) << endl;);
    return sub_path(obj, Loc(0,0), Loc(-1,-1), goal_obj, output, distance, lod, consider, stop_time);
  }

  //-----------------------------------------------------------------------------
  bool ST_Terrain::PFEngine::sub_path(const Object *obj, 
                                      const Loc &topLeft, 
                                      const Loc &bottomRight, 
                                      const Loc &goal_loc, 
                                      Path &output, 
                                      sint4 distance,
                                      LevelOfDetail lod,
                                      ConsiderObjects consider,
				      uint4 stop_time)
  {
    // Use the appropriate 'goal tester'
    IsNearLocation tester(goal_loc, distance);
    bool found = find_path(obj, get_center_loc(obj), goal_loc, tester,
                           topLeft, bottomRight, lod, consider, output, stop_time);


    /** The next refinement doesn't work very well:
        If the target location is very near to an obstacle, such that the size
        of the moving object forbids it from being there, the PF will search the 
        entire map (which is slow) before realizing this. */
#if 0
    /* To overcome an inaccuracy of our internal low resolution, if
       the end-of-path location (expressed in world coords) is not
       exactly at the goal location, do a final PF using the max
       resolution.
       fixme: if it's too slow, use a constrained submap to prevent full-map searches. */
    if( found && ! tester.test_for_world_coords(output.locs[0].x, output.locs[0].y) ) {
      // If there is a possibility for improvement (+prevent an infinite recursion)
      if( lod != max_details_LOD ) { 

        // Plan the path from the last waypoint found
        Path final_leg;
        bool found_final_leg = find_path(obj, output.locs[0], goal_loc, tester,
                                         topLeft, bottomRight, max_details_LOD, 
                                         consider, final_leg, stop_time);

        if( found_final_leg )
          output.locs.insert(output.locs.begin(), final_leg.locs.begin(), final_leg.locs.end());
      }
    }
#endif
    return found;
  }

  //-----------------------------------------------------------------------------
  bool ST_Terrain::PFEngine::sub_path(const Object *obj, 
                                      const Loc &topLeft, 
                                      const Loc &bottomRight, 
                                      const Object *goal_obj, 
                                      Path &output, 
                                      sint4 distance,
                                      LevelOfDetail lod,
                                      ConsiderObjects consider,
				      uint4 stop_time)
  {
    Loc goal_loc; 
    goal_obj->get_center(goal_loc.x, goal_loc.y);
    IsNearObject tester(obj, goal_obj, distance);
    bool found = find_path(obj, get_center_loc(obj), goal_loc, tester, 
                           topLeft, bottomRight, lod, consider, output, stop_time);

    /* To overcome an inaccuracy of our internal low resolution, if
       the end-of-path location (expressed in world coords) is not
       close enough to the goal object, do a final PF using the max
       resolution.
       fixme: if it's too slow, use a constrained submap to prevent full-map searches. */
    if( found && ! tester.test_for_world_coords(output.locs[0].x, output.locs[0].y) ) {
      // If there is a possibility for improvement (+prevent an infinite recursion)
      if( lod != max_details_LOD ) { 

        // Plan the path from the last waypoint found
        Path final_leg;
        bool found_final_leg = find_path(obj, output.locs[0], goal_loc, tester,
                                         topLeft, bottomRight, max_details_LOD,
                                         consider, final_leg, stop_time);

        if( found_final_leg )
          output.locs.insert(output.locs.begin(), final_leg.locs.begin(), final_leg.locs.end());
      }
    }

    return found;
  }

  //-----------------------------------------------------------------------------
  template <class GoalTester>
    bool ST_Terrain::PFEngine::find_path(const Object *obj,
					 const Loc &start_position,
					 const Loc &goal,
					 const GoalTester &goal_tester,
					 const Loc &topleft,
					 const Loc &bottomright,
					 LevelOfDetail lod,
					 ConsiderObjects consider,
					 Path& output,
					 const ControlParams &control_params)
  {
    assert(lod < lods.size());

    // select the appropriate map
    shared_ptr<Map> m;
    if( obj->get_zcat() == Object::ON_LAND )
      m = ground_maps[lod];
    else
      m = air_maps[lod];
    //MapPtr m = (obj->get_zcat() == Object::ON_LAND)? &map : &air_map;

    // adjust submap limits for the case of 'use the entire map'
    Loc br = bottomright;
    if( br == Loc(-1, -1) )
      br = Loc(m->get_w()-1, m->get_h()-1);
    assert(topleft.x >= 0 && topleft.x <= br.x && br.x < m->get_w() &&
           topleft.y >= 0 && topleft.y <= br.y && br.y < m->get_h() && "Invalid submap coordinates");
    
    // remove objects we wish to ignore
    ignore_objects(consider, m, obj);

    // When working with a low resolution, it might happen that the outer radius
    // of the object occupies cells that are also used by other objects. 
    // If this happens, the first nodes considered by A* will have a very 
    // high cost. 
    // To prevent the search from failing, we start by going to a near
    // location (in the direction of the original goal) that is 'clear', 
    // ie have a low cost - using the full resolution map.

    // fixme: Now that we have ControlParams in place, it might be possible to 
    // clear directly using the low-res. Consider this option.

    Path clear;
    if( lod != max_details_LOD ) {
      // don't fail when expanding high-cost nodes since we might be 
      // located in such an area to begin with.
      ControlParams cp(control_params.stop_time, 0);
      // get to a clear location.
      bool found = find_path(obj, start_position, goal, IsAtClearLocation(m, obj),
                             topleft, bottomright, max_details_LOD,
                             consider, clear, cp);
      if( !found ) {
        // put the ignored objects back into the world map before returning
        reconsider_objects(consider, m, obj);
        return false;
      }
    } else {
      // start from the given position
      clear.locs.push_back(start_position);
    }

    // start from the clear position:
    sint4 x1 = m->world2x(clear.locs[0].x);
    sint4 y1 = m->world2y(clear.locs[0].y);

    // the goal position for the heuristic in A*
    sint4 x2 = m->world2x(goal.x);
    sint4 y2 = m->world2y(goal.y);

    // constrain goal to inside the submap
    sint4 r = m->world2r(obj->get_radius());
    x2 = min(max(x2,topleft.x+r+1), br.x-r-1);
    y2 = min(max(y2,topleft.y+r+1), br.y-r-1);

    // a bucket representing the moving object
    Bucket b = BucketFactory::get_bucket(r);
    b.set_center(Loc(x1, y1));

    // now find a path
    //cout << "FINDING PATHS" << endl;
    uint4 end_of_path_cell;
    AStarMap &astar_map = *(astar_maps[lod]);
    bool found = simple_astar(*m, astar_map, b, x2, y2, goal_tester, 
                              topleft, br, end_of_path_cell, control_params);

    // and return the ignored objects to the world map
    reconsider_objects(consider, m, obj);

    if(found){
      // walk back from the goal, listing all waypoints
      uint4 waypoint = end_of_path_cell;
      do {
        //map(waypoint).flags = 1;
        output.locs.push_back(Loc(m->x2world(m->i2x(waypoint)), 
                           m->y2world(m->i2y(waypoint))));
        waypoint = astar_map.at(waypoint).parent;
      } while(waypoint);

      // add the clearing portion. It ends where output.locs starts, so
      // drop this duplicate waypoint (remember waypoints are listed in reverse order).
      output.locs.insert(output.locs.end(), clear.locs.begin()+1, clear.locs.end());

      // if there is more than one waypoint ==> start!=end of path
      if (output.locs.size() >= 2) {

        // smooth the path
        smoothen_path(output.locs);

        // remove the first "goal" which is the current==start location      
        output.locs.pop_back();
      }
    }

    //output.id = ? // fixme: output.id should be a PathId
    return found;
  }

  //-----------------------------------------------------------------------------
  template<class GoalTester>
    bool ST_Terrain::PFEngine::simple_astar(const Map &smap, 
                                            AStarMap &astar_map,
                                            Bucket b, 
                                            sint4 x2, sint4 y2,
                                            const GoalTester &goal_tester,
                                            const Loc &topleft,
                                            const Loc &bottomright,
                                            uint4 &end_of_path,
					    const ControlParams &control_params)
  {
    TEST(cout << "Current time=" << SDL_GetTicks() << " and this A* call has to to stop until " << control_params.stop_time << endl;);

    // distances to the eight neighbour cells, starting from east, southeast...
    static const uint4 cost[8] = {FACTOR, DIAG, FACTOR, DIAG, FACTOR, DIAG, FACTOR, DIAG};

    // increment astar_map's magic number
    uint4 magic = astar_map.get_magic() + 1;
    if (magic >= 1000000000) magic = 1; // wrap around
    astar_map.set_magic(magic);
    
    // reset open/closed set membership variables when "magic" is 1 after increment
    if (magic == 1) {
      FORS (i, astar_map.get_n()) astar_map(i).open = astar_map(i).closed = 0;
    }

    sint4 x1 = b.get_center().x;
    sint4 y1 = b.get_center().y;

    priority_queue<HeapElement, vector<HeapElement>, HeapElementCmp> open;
    AStarCell &s = astar_map(x1, y1);

    //  assert(size >= 1);
    //  assert(smap(x1, y1).max_size >= size);

    //  if (smap(x2, y2).max_size < size) return false; // goal not reachable

    HeapElement hs;
    hs.cell = astar_map.index(x1, y1);
    hs.g = 0;
    hs.f = estimate(x1, y1, x2, y2);
    open.push(hs);

    //cout << "est=" << hs.f << endl;

    s.parent = 0; // no predecessor
    s.open = magic;
    s.g = 0;

    size_t max_heap = 0;
    uint4 count = 0;

    TEST2( cout << "A* expanded nodes:" << endl; );

/*    // Debugging aid:
    bool to_file = false;
    ofstream file;
    if( to_file )
      file.open("T.txt");
    sint4 testx=-1, testy=-1;*/

    while( !open.empty() ) {
      ++count;
      HeapElement he = open.top(); open.pop();

      // skip if he no longer up-to-date
      if (astar_map(he.cell).g != he.g) 
        continue;

      // skip out-of-submap nodes
      if (astar_map.i2x(he.cell) < topleft.x || astar_map.i2x(he.cell) > bottomright.x ||
          astar_map.i2y(he.cell) < topleft.y || astar_map.i2y(he.cell) > bottomright.y )
          continue;

      TEST2( cout << astar_map.i2x(he.cell) << "," << astar_map.i2y(he.cell) << ";" << endl; );

/*      // Debugging aids:
      if( to_file )
        file << astar_map.i2x(he.cell) << "," << astar_map.i2y(he.cell) << endl;
      if( astar_map.i2x(he.cell) == testx && astar_map.i2y(he.cell) == testy )
        cout << "!" << endl;*/

      // expand goal => path found
      if( goal_tester(astar_map.i2x(he.cell), astar_map.i2y(he.cell), smap) ) { 
        end_of_path = he.cell;
        return true;
      }

      // check stop-and-fail criterion: time
      if( control_params.stop_time != 0 && SDL_GetTicks() >= control_params.stop_time ) {
        TEST(cout << "Ran out of time for this A* search." << endl;);
        return false;
      }

      // check stop-and-fail criterion: high cost
      // (This isn't necessarily the best thing to do. We can try and move there
      //  and allow the collision. For example, before the "move to touch object"
      //  machinery was created, this was the only way to move to an object.)
      if( control_params.stop_cost != 0 && astar_map(he.cell).g >= control_params.stop_cost ) {
        TEST(cout << "Expanded an occupied node, no path exists." << endl;);
        return false;
      }

      // Expand this node
      // ----------------

      b.set_center(Loc(smap.i2x(he.cell), smap.i2y(he.cell)));

      HeapElement hen;

      // Check all directions
      bitset<8> freeDirection;
      boost::array<uint4, 8> moveCosts;
      b.check_all_directions(smap, freeDirection, moveCosts);

      FORS (i, 8) {
        //      if (i%2) continue;

        //        sint4 moveCost = 0;
        //        if (!b.check_direction(smap, i, moveCost)) continue;
        if (!freeDirection[i]) continue;

        uint4 nn_index = he.cell + smap.get_d(i); // neighbour

        // cout << int(nn->max_size) << " " << int(size) << endl;

        AStarCell &nn = astar_map(nn_index);
        //        uint4 newg = he.g + cost[i] + moveCost;
        uint4 newg = he.g + cost[i] + moveCosts[i];

        if ((nn.open == magic || nn.closed == magic) && nn.g <= newg) continue;

        nn.parent = he.cell;
        nn.g = newg;

        hen.cell = nn_index;
        hen.g = newg;
        hen.f = newg + estimate(smap.i2x(nn_index), smap.i2y(nn_index), x2, y2);

#if 0
        cout << "new g=" << hen.g
          << " h=" << estimate(smap.i2x(nn_index), smap.i2y(nn_index), x2, y2)
          << " f=" << hen.f
          << endl;
#endif

        if (nn.closed == magic) nn.closed = 0;
        open.push(hen); nn.open = magic;
        max_heap = max(open.size(), max_heap);
      }

      astar_map(he.cell).closed = magic;
    }

    TEST(cout << "Checked every open node and couldn't find a path!" << endl;);
    return false;
  }

  //-----------------------------------------------------------------------------
  // Changes here should be reflected in reconsider_objects()
  void ST_Terrain::PFEngine::ignore_objects(ConsiderObjects consider, 
                                            shared_ptr<Map> map, 
                                            const Object *moved_obj)
  {
    if( consider & IGNORE_ALL_MOBILE_OBJS ) {
      FORALL(mobile_objects, obj)
        map->remove_object(*obj);
    }
    else {
      if( consider & IGNORE_MOBILE_FRIENDS ) {
        ERR("IGNORE_MOBILE_FRIENDS still unimplemented");
      }
      
      if( consider & IGNORE_MOBILE_FOES ) {
        ERR("IGNORE_MOBILE_FOES still unimplemented");
      }

      if( consider & IGNORE_MOBILE_NEUTRALS ) {
        ERR("IGNORE_MOBILE_NEUTRALS still unimplemented");
      }
    }
    // If the moved_obj is in the world representation, and we didn't
    // yet remove it as part of all friendly mobiles, remove it now so
    // it won't interfere to planning its own path.
    if( ~consider & IGNORE_MOBILE_FRIENDS ) {
      FIND(mobile_objects, i, moved_obj);
      if( i != mobile_objects.end() )
        map->remove_object(moved_obj);
    }
  }

  //-----------------------------------------------------------------------------
  // Changes here should be reflected in ignore_objects()
  void ST_Terrain::PFEngine::reconsider_objects(ConsiderObjects consider, 
                                                shared_ptr<Map> map, 
                                                const Object *moved_obj)
  {
    if( consider & IGNORE_ALL_MOBILE_OBJS ) {
      FORALL(mobile_objects, obj)
        map->insert_object(*obj);
    }
    else {
      if( consider & IGNORE_MOBILE_FRIENDS ) {
        ERR("IGNORE_MOBILE_FRIENDS still unimplemented");
      }
      
      if( consider & IGNORE_MOBILE_FOES ) {
        ERR("IGNORE_MOBILE_FOES still unimplemented");
      }

      if( consider & IGNORE_MOBILE_NEUTRALS ) {
        ERR("IGNORE_MOBILE_NEUTRALS still unimplemented");
      }
    }

    // If the moved_obj was in the world representation, and we didn't
    // yet returned it as part of all friendly mobiles, put it back in now
    if( ~consider & IGNORE_MOBILE_FRIENDS ) {
      FIND(mobile_objects, i, moved_obj);
      if( i != mobile_objects.end() )
        map->insert_object(moved_obj);
    }
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::ignore_rectangles()
  {
    MapsVector::iterator map = ground_maps.begin();
    for(; map != ground_maps.end() ; map++) {
      FORALL(rectangle_objects, obj)
        (*map)->remove_object(*obj);
    }
  }

  void ST_Terrain::PFEngine::consider_rectangles()
  {
    MapsVector::iterator map = ground_maps.begin();
    for(; map != ground_maps.end() ; map++) {
      FORALL(rectangle_objects, obj)
        (*map)->insert_object(*obj);
    }
  }

  //-----------------------------------------------------------------------------
  /**
  * go throught the waypoints in the given vector and remove
  * all redundent points that fall on a straight line
  * */
  void ST_Terrain::PFEngine::smoothen_path(Vector<Loc> &path)
  {
    uint4 i1 = 0;
    uint4 i2 = 1;

    for (uint4 i = 2; i < path.size(); i++) {

      while (i < path.size() && on_line(path[i1], path[i2], path[i])) {
        path.erase(path.begin()+i2);
      }

      i1 = i-1;
      i2 = i;
    }
  }

  //-----------------------------------------------------------------------------
  uint4 ST_Terrain::PFEngine::estimate(sint4 x1, sint4 y1, sint4 x2, sint4 y2)
  {
    /*
    *  -----
    *       \   
    *        \ 
    *
    * Octile distance: 
    *   dmin * sqrt(2) + dmax - dmin
    */

    sint4 dx = abs(x1-x2);
    sint4 dy = abs(y1-y2);
    sint4 dmin = min(dx, dy);
    return dmin * DIAG + (dx + dy - dmin - dmin)*FACTOR;

    // Euclidean
    //return sint4(sqrt(dx*dx + dy*dy)*FACTOR);

    // Manhattan
    // return max(dx, dy)*FACTOR;
  }

  //-----------------------------------------------------------------------------
  bool ST_Terrain::PFEngine::on_line(const Loc &a, const Loc &b, const Loc &c)
  {
    if (a.x == b.x && b.x == c.x) { return true; }
    if (a.y == b.y && b.y == c.y) { return true; }
    double tx = (double)(b.x - a.x) / ((double)(c.x - a.x));
    double ty = (double)(b.y - a.y) / ((double)(c.y - a.y));
    return tx >=0 && tx <= 1 && tx == ty;
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::insert_object(const Object *obj)
  {
    // insert into all relevant maps
    MapsVector &maps = (obj->get_zcat() == Object::ON_LAND)? ground_maps : air_maps;
    for(MapsVector::iterator m = maps.begin() ; m != maps.end() ; m++)
      (*m)->insert_object(obj);
    // if it's mobile, remember for future reference
    if( obj->get_max_speed() > 0 )
      mobile_objects.insert(obj);
    else {
      // if it's a non-mobile rectangle, remember for future reference
      if( obj->get_shape() == Object::RECTANGLE )
        rectangle_objects.insert(obj);
    }
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::remove_object(const Object *obj)
  {
    // remove from all relevant maps
    MapsVector &maps = (obj->get_zcat() == Object::ON_LAND)? ground_maps : air_maps;
    for(MapsVector::iterator m = maps.begin() ; m != maps.end() ; m++)
      (*m)->remove_object(obj);
    // if needed, remove from the set of mobile objects
    if( obj->get_max_speed() > 0 )
      mobile_objects.erase(obj);
    else {
      // or from the set of rectangles
      if( obj->get_shape() == Object::RECTANGLE )
        rectangle_objects.erase(obj);
    }  
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::update_object(const Object *obj)
  {
    MapsVector &maps = (obj->get_zcat() == Object::ON_LAND)? ground_maps : air_maps;
    for(MapsVector::iterator m = maps.begin() ; m != maps.end() ; m++)
      (*m)->update_object(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::insert_boundary(const Segment& l)
  {
    // insert to all ground maps (boundaries are currently only relevant on land)
    for(MapsVector::iterator m = ground_maps.begin() ; m != ground_maps.end() ; m++)
      (*m)->insert_boundary(l);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::ir_path(const Object *obj, const Path &path, bool insert, bool with_endpoints)
  {
    // insert (/remove) to all ground maps
    for(MapsVector::iterator m = ground_maps.begin() ; m != ground_maps.end() ; m++)
      (*m)->ir_path(obj, path, insert, with_endpoints);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::init_sizes(SimpleMap<SizeCell> &smap)
  {
    FORU(i, smap.get_n()) { 
      smap(i).max_size = 1; 
      //smap(i).flags = OPEN; 
    }
    const sint4 w = smap.get_w(), h = smap.get_h();  
    // mark border as non-traversable
    FORS (x, w) { 
      smap(x, 0).max_size = 0; smap(x, h-1).max_size = 0;
      //smap(x, 0).flags = CLOSED; smap(x, h-1).flags = CLOSED;
    }
    FORS (y, h) { 
      smap(0, y).max_size = 0; smap(w-1, y).max_size = 0; 
      //smap(0, y).flags = CLOSED; smap(w-1, y).flags = CLOSED;
    }
  }

  //-----------------------------------------------------------------------------
  //-----------------------------------------------------------------------------
  //-----------------------------------------------------------------------------

  void ST_Terrain::PFEngine::Map::insert_object(const Object *obj)
  {
    ir_object(obj, true);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::Map::remove_object(const Object *obj)
  {
    ir_object(obj, false);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::Map::update_object(const Object *obj)
  {
    // remove previous values
    ir_object(obj, false, true);
    // insert new values
    ir_object(obj, true, false);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::Map::insert_boundary(const Segment& l)
  {
    bool costonly;
    Options::get("-costonly", costonly);
    if (costonly) {
      insert_line(world2x(l.l1.x), world2x(l.l1.y), world2x(l.l2.x), world2x(l.l2.y), 
                  0, BOUND_COST);
    } else {
      insert_line(world2x(l.l1.x), world2x(l.l1.y), world2x(l.l2.x), world2x(l.l2.y), 
                  1, 0);
    }
  }

  //-----------------------------------------------------------------------------
  struct LegMotionData {
    static const sint4 fine_grid_res_bits = 6;
    ScalarPoint start_f_pos, target_f_pos, f_delta;
    real8 inv_ttt;
    uint4 ticks_total;
    sint4 speed;

    LegMotionData(const Loc &spos, const Loc &tpos, sint4 speed_)
      : speed(speed_)
    {
      start_f_pos.x  = spos.x << fine_grid_res_bits;
      start_f_pos.y  = spos.y << fine_grid_res_bits;
      target_f_pos.x = tpos.x << fine_grid_res_bits;
      target_f_pos.y = tpos.y << fine_grid_res_bits;

      // vector to the target
      f_delta = ScalarPoint(target_f_pos.x - start_f_pos.x, target_f_pos.y - start_f_pos.y);
      Scalar dist = f_delta.l2_norm();
      //real8 dist = start_f_pos.distance(target_f_pos);

      // 1 / time to reach target
      inv_ttt = real8(speed << fine_grid_res_bits)/dist;
      ticks_total = sint4(ceil(1.0/inv_ttt));

#if 0
      if (dist == 0) {
        obj_stop(e);
        return false;
      }
      if (delta.x == 0 && delta.y == 0) return false;
      if (e.smd.ticks_total <= 0) ERR("ticks <= 0");
#endif
    }

    Loc location_at_tick(uint4 tick) const {
      // fraction of distance travelled to the target
      real8 s = inv_ttt * tick;
      ScalarPoint pos;
      if (s >= 1.0) {
        pos = target_f_pos;
      } else {
        pos.x = start_f_pos.x + Scalar(s * f_delta.x);
        pos.y = start_f_pos.y + Scalar(s * f_delta.y);
      }
      
      pos.round(fine_grid_res_bits);
      pos.x >>= fine_grid_res_bits;
      pos.y >>= fine_grid_res_bits;
      return Loc(pos.x, pos.y);
    }
  };

  struct null_deleter { void operator()(void const *) const {} };

  // Go over the path and insert/remove the object at each location
  void ST_Terrain::PFEngine::Map::ir_path(const Object *obj, const Path &path, bool insert, bool with_endpoints)
  {
    // The test for keeping the start/end points unblocked is implemented in a very
    // primitive way. Rethink someday.

    ObjectTranslocator o(obj);
    shared_ptr<Map> p(this, null_deleter());

    Loc prev_waypoint;
    obj->get_center(prev_waypoint.x, prev_waypoint.y);

    // for keeping the start/end points clear, if it was requested
    Loc startpoint = prev_waypoint;
    Bucket start_bucket = BucketFactory::get_bucket(world2r(obj->get_radius()));
    start_bucket.set_center(Loc(world2x(startpoint.x), world2y(startpoint.y)));

    Loc endpoint   = path.locs.back();
    Bucket end_bucket(start_bucket);
    end_bucket.set_center(Loc(world2x(endpoint.x), world2y(endpoint.y)));

    FORALL(path.locs, waypoint) {

      LegMotionData lmd(prev_waypoint, *waypoint, 1);
      uint4 tick = (waypoint == path.locs.begin())? 0:1;
      for( ; tick <= lmd.ticks_total ; tick++ ) {
        Loc loc = lmd.location_at_tick(tick);
        o.set_position(loc.x, loc.y);

        // if we don't want to block the start/end points, check before the 
        // insertion/removal that this will not overlap the endpoints
        if( !with_endpoints ) {
          uint4 start_cost_before = 0, start_cost_after = 0;    
          uint4 end_cost_before = 0, end_cost_after = 0;    

          start_bucket.check_current_location(*this, start_cost_before);
          end_bucket.check_current_location(*this, end_cost_before);
          ir_object(&o, true);
          start_bucket.check_current_location(*this, start_cost_after);
          end_bucket.check_current_location(*this, end_cost_after);
          ir_object(&o, false);

          // if overlapping the end, then we finished
          if( end_cost_before != end_cost_after )
            break;
          // if overlapping the start, then skip this loc
          if( start_cost_before != start_cost_after )
            continue;
        }

        // insert/remove at this loc
        ir_object(&o, insert);
      }
      prev_waypoint = *waypoint;
    }
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::Map::ir_object(const Object *obj, 
                                            bool insert, 
                                            bool use_prev_attributes)
  {
    sint4 cost = insert ? 1 : -1;

    if (obj->get_shape() == Object::RECTANGLE) {
      //CellFlag f = (insert)? TEMP : OPEN;
      Loc topleft, bottomright;

      if( !use_prev_attributes ) {
        obj->get_p1(topleft.x, topleft.y);
        obj->get_p2(bottomright.x, bottomright.y);
      } else {
        // get previous attribute values
        sint4 dummy1, dummy2;
        obj->get_p1(dummy1, dummy2, &(topleft.x), &(topleft.y));
        obj->get_p2(dummy1, dummy2, &(bottomright.x), &(bottomright.y));
      }

      insert_rectangle(topleft, bottomright, /*f, 0*cost,*/ cost*UNIT_COST);

    } else if (obj->get_shape() == Object::CIRCLE) {

      sint4 r = obj->get_radius();
      Loc worldc;

      if( !use_prev_attributes ) {
        obj->get_center(worldc.x, worldc.y);
      } else {
        // get previous attribute values
        sint4 dummy1, dummy2;
        obj->get_center(dummy1, dummy2, &(worldc.x), &(worldc.y));
      }

      int newr = world2r(r);
      Bucket b = BucketFactory::get_bucket(newr);
      Loc center(world2x(worldc.x), world2y(worldc.y));

      b.set_center(center);
      insert_bucket(b, 0, cost*UNIT_COST);

      if( newr > 0 ) {
        Bucket b2 = BucketFactory::get_bucket(newr-1);
        b2.set_center(center);
        insert_bucket(b2, 0, cost*UNIT_COST/2);
        }

      // fixme: heading is not accessible through Object
      /* We wanted to add some cost to the position where the object will be
      in the next tick to prefer not to go there; However, heading is not
      currently accessible with the new TerrainBase interface:
      // one second distance
      Loc head = cworld;
      Loc tail = cworld;
      get_head_and_tail(head, tail, heading);

      b.set_center(head);
      insert_bucket(b, zcat, TEMP, 0, id, cost*CROSS_COST);

      b.set_center(tail);
      //  insert_bucket(b, zcat, TEMP, 0, id, cost*CROSS_COST);
      */
    } else ERR("unhandled shape");
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::Map::insert_rectangle(const Loc &topleft, const Loc &bottomright, sint4 cost)
  {
    sint4 x1 = world2x(topleft.x);
    sint4 y1 = world2y(topleft.y);
    sint4 x2 = world2x(bottomright.x);
    sint4 y2 = world2y(bottomright.y);
#if 0
    insert_line(x1, y1, x1, y2, 0, cost);
    insert_line(x1, y2, x2, y2, 0, cost);
    insert_line(x2, y2, x2, y1, 0, cost);
    insert_line(x2, y1, x1, y1, 0, cost);
#endif
    // fill the interior to prevent moves into it when using low resolution representations.
    for(sint4 y = y1 ; y < y2 ; y++ )
      insert_line(x1, y, x2, y, 0, cost);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::Map::insert_bucket(const Bucket &b, sint4 delta_maxsize, sint4 cost)
  {
    sint4 x1, y1;
    const Loc &c = b.get_center();
    const Vector<Loc> &points = b.get_points();

    FORALL (points, i) {
      x1 = c.x + i->x;
      y1 = c.y + i->y;
      set_cell(x1, y1, delta_maxsize, cost);
    }
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::PFEngine::Map::insert_line(sint4 x1, sint4 y1, sint4 x2, sint4 y2, 
                                              sint4 delta_maxsize, sint4 cost)
  {
    int dx = x2-x1;
    int dy = y2-y1;
    
    if (dx == 0) {
    if (y1 < y2) FORT (i, y2-y1+1) set_cell(x1, y1+i, delta_maxsize, cost); 
    else         FORT (i, y1-y2+1) set_cell(x1, y2+i, delta_maxsize, cost);
    return;
    }

    if (dy == 0) {
    if (x1 < x2) FORT (i, x2-x1+1) set_cell(x1 + i, y1, delta_maxsize, cost);
    else         FORT (i, x1-x2+1) set_cell(x2 + i, y1, delta_maxsize, cost);
    return;
    }

    if (dx == dy) {
    FORT (i, x2-x1)   set_cell(x1+i+1, y1+i,   delta_maxsize, cost);
    FORT (i, x2-x1+1) set_cell(x1+i,   y1+i,   delta_maxsize, cost);
    FORT (i, x2-x1)   set_cell(x1+i,   y1+i+1, delta_maxsize, cost);
    return;
    }

    if (dx == -dy) {
    FORT (i, x2 - x1    ) set_cell(x1+i  , y1-i-1, delta_maxsize, cost);
    FORT (i, x2 - x1 + 1) set_cell(x1+i  , y1-i  , delta_maxsize, cost);
    FORT (i, x2 - x1    ) set_cell(x1+i+1, y1-i  , delta_maxsize, cost);
    return;
    }
    
    ERR("can't insert general line");
  }

  //-----------------------------------------------------------------------------
  /**
  * set the info stored in the cell at x,y in map m to given information
  * @param m the map to be edited
  * @param x,y the x and y coordinates of the cell
  * @param f flag to set the cell to
  * @param size amount by which the size paramiter of the cell is _changed_
  * @param cost amount by which the cost paramiter of the cell is _changed_
  * */
  void ST_Terrain::PFEngine::Map::set_cell(sint4 x, sint4 y, sint4 delta_maxsize, sint4 cost)
  {
    if (x < 0 || x >= get_w() || y >= get_h() || y < 0  ) {
      //    cout << "invalid coodinates : " << "[" <<x << "," << y << "]" << endl;
      return;
    }

    SizeCell &c = operator()(x, y);

    if (c.max_size > 0 && delta_maxsize > 0 ) {
      c.max_size = 0;
    } else {
      c.max_size -= delta_maxsize;
    }

    //c.flags  = f;
/*    if( (cost < 0 && ((sint8)c.cost < -(sint8)cost)) )
      cout << "!" << endl;*/
    assert((cost >= 0 || ((sint8)c.cost >= -(sint8)cost)) && "Cannot reduce cost below 0");
      c.cost += cost; // if (c->cost < 0) c->cost = 0;
/*    if( c.cost > 1000000000 )
      cout << "!" << endl;*/
    assert( c.cost <= 1000000000 ); 
  }

  //-----------------------------------------------------------------------------
/*  bool ST_Terrain::PFEngine::Map::is_blocked(const Object *obj, const Loc &world_location)
  {
    assert(obj->get_shape() == Object::CIRCLE);

    sint4 r = obj->get_radius();
    int newr = world2r(r);
    Bucket b = BucketFactory::get_bucket(newr);
    Loc center(world2x(world_location.x), world2y(world_location.y));
    b.set_center(center);
    
    insert_bucket(b, 0, cost*UNIT_COST);

  }
*/

} // End of namespace SimpleTerrain

