#ifndef TMANAGER_H
#define TMANAGER_H

/* Headerfile for tournament manager software.
 * (updated)
 * June 9, 2006
 */

// $Id: tmanager.h,v 1.1.1.1 2007/04/09 19:06:55 pjungkun Exp $

#include <pthread.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <memory.h>    
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <string>
#include <fstream>
#include <sstream>
#include <istream>
#include <iostream>
#include <algorithm>
#include "Vector.H"

using namespace std;

#define true      1
#define false     0
#define FOR(i,n)  for (i=0; i < (n); i++)
#define FORS(i,n)  for (int i=0; i < (n); i++)
#define FOREVER   for (;;)

// assumes container does not change
#define FORALL(CONT,i)  \
  for (typeof((CONT).begin()) i = (CONT).begin(), _ = (CONT).end(); i != _; ++i) 

// assumes container does not change and defines iterator outside of loop (B=BREAK) 
// ATTENTION: two lines of code => { } may be necessary!
#define FORALL_B(CONT,i) \
  typeof((CONT).begin()) i = (CONT).begin(); \
  for (typeof((CONT).begin()) _ = (CONT).end(); i != _; ++i) \


#define NO_RES 0        //For tournament matrix display: 
                        //0-only display result, 1- display everything other than result


enum result_flag_t { queued, pending, done };

struct Player
{
  string name;     // player name
  string username; // player's user name 
  int index;       // player index
  string startLine;// startline of the client's executable(game type is appended to it)  
  string workDir;  // client working directory
  double score;    // player's running score
  int maxPoints;   // total possible score for a player (# matches played so far)
  int errors;      // number of games with errors

  Player()
  {
    index = -1;
    score = 0;
    maxPoints = 0;
    errors = 0;
  }
};

struct Computer
{
  string name;     // computer name
  int index;       // computer index
  int taskCount;   // number of processes on this computer (can have
		   // more than 1 process on a server)
  bool isServer;   // true <=> computer is a server
  bool isUp;       // true <=> computer is up and ok to use
  
  int portNum;     // next available port number for a server on this computer
  
  Computer()
  {
    index = -1;
    taskCount = -1;
    isServer = isUp = false;
    portNum = 7777;
  }
};

struct Game
{       
  int id;                    // game index
  int gameType;              // type of the game (1,2,or 3)
  result_flag_t progress;    // indicates whether the game is queued, pending, or done 
  Vector<string> result;     //[3] result strings -- 0 is server, 1 is
			     // player1, 2 is player2
  int seed;                  // game (map) seed
  Vector<int> player_list;   //[2] indices of players of the game): 1
			     // element for game type 1,
			     // elements for types 2 & 3
  Vector<Computer*> clients; // client computers for this game: 1
			     // client for game type 1, 2 for types 2 & 3
  Computer *server;          // the server computer for this game
  int port;                  // port to connect to (could be multiple servers running)
  timeval start_time;        // timestamp of the start of this game              
  Vector<timeval> end_times; //[3] timestamp of the end of each thread 
  double maxTime;            // the max allowable time for a game
			     // (including a grace period)
  int tries;                 // # of times this game was run so far
  bool saved;                // true <=> this game's results are saved in the config file
  pthread_t thread_game;     // game thread  
  pthread_t thread_srv;      // server thread
  Vector<pthread_t> threads_ply; //[2] player threads
  Vector<result_flag_t> thread_status; //[3] pending at the beginning
				       // of the thread, done at the end
  
  string serverStartLine;    // append game id to it
  string serverWorkDir;      // working directory
  string serverUserName;     // username of server

  string serverOutDir;       // in what subdirectory server output is stored
  string replayDir;          // replays are stored here
  string clientOutDir;       // in what subdirectory client output is stored
  
  unsigned int magic;

  enum { MAGIC = 0xdeadbeaf };
    
  Game()
    : result(3), player_list(2), clients(2),
      end_times(3), threads_ply(2), thread_status(3)
  {
    magic = MAGIC;
    id = gameType = 0;
    progress = done;
    seed = 0;
    clients[0] = clients[1] = 0;
    server = 0;
    port = 0;
    maxTime = 0;
    tries = 0;
    saved = false;
    thread_srv = thread_game = threads_ply[0] = threads_ply[1] = 0;
    thread_status[0] = thread_status[1] = thread_status[2] = done;
  }
};

struct GameArg
{
  Game *game;
  int playerIndex;

  GameArg() { game = 0; playerIndex = -1; }
};

struct Summary
{
  int player_list[2];
  float scores[2];
  int matches;
  
  Summary() { 
    player_list[0] = player_list[1] = 0;
    scores[0] = scores[1] = 0;
    matches = 0;
  }
};

// global game data structures

extern Vector<Game>     games; 
extern Vector<Computer> computers;  
extern Vector<Player>   players;    

extern pthread_mutex_t mutexGames;    


// Helper Functions:

/**
 * Error reporting function, which displays the given error
 * message to standard error and terminates the program with code 20.
 *
 * @param msg  error message to be displayed
 */
void Error(string msg);

/**
 * Helper function to get the current time using gettimeofday.
 *
 * @return timeval type which contains the timing information
 */
timeval get_time();

/**
 * Helper function to get the number of seconds elapsed
 * since 1/1/1970.
 *
 * @param ti  timeval parameter containing the timing information
 * @return    time in seconds since 1/1/1970
 */
double to_double(const timeval &ti);

/**
 * Helper function for parsing the configuration file which
 * skips sections of comments (denoted by #).
 * 
 * @param in  the input file stream (configuration file)
 * @return    a bool flag indicating whether or not end-of-file is reached (true - eof, false - otherwise)
 */
bool skipComments(istream &in);

/**
 * Helper function for converting an int to a string.
 *
 * @param i   integer to be converted to a string
 * @return    string version of the integer
 */
string intToString(int i);

/**
 * Helper function for converting a string representing
 * a floating point number to a double.
 *
 * @param s  the string to be converted to a double
 * @return   the double value represented by the string parameter
 */
double stringToNum(string s);

/**
 * Helper function to scan server output string to 
 * check for client disconnect (<newline><id>:<spaces>dead).
 *
 * @param  src  server output string to be scanned
 * @param  id   0 or 1 depending on which player is being checked for disconnect
 * @return      true <=> this player was disconnected  
 */
bool playerDisconnected(string src, int id);

/**
 * Displays the help message to standard out when -help is the command line argument.
 * Exits the program with code 10.
 */
void print_help();

/**
 * Displays all games in the queue. Shows game index, game seed,
 * player1 name, player2 name
 *
 * @param gameType   the type of game(1,2,or3)
 */
void displayGames(int gameType);

/**
 * Displays game statistics for all players.
 *
 */
void displayScores();

//Central Functions:

/**
 * Generates the game queue.
 * 
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 * @param game_seed       base map seed found by XORing all players' seeds
 * @param gameType        the type of game (1, 2, or 3)
 * @param game_length     the expected length of game 
 * @param srvStartLine    startline of the server's executable (game type will be appended to it)
 * @param srvWorkDir      working directory of the server
 * @param srvUsrName      user name of the server
 * @param srvOutDir       server output subdirectory
 * @param replayDir       replay subdirectory
 * @param cliOutDir       client output subdirectory
 * @return                total number of games in game_queue
 */
void buildGameList(int games_per_pair, bool single, int game_seed,
		   int gameType, double game_length, string srvStartLine,
		   string srvWorkDir, string srvUsrName,
		   string srvOutDir, string replayDir, string cliOutDir);

/**
 * Makes accessible both the standard input and output of the shell 
 * process it creates. Two pipes are created: one provides standard
 * input for the shell commmand, and the other for passing back its 
 * standard output. The pipe file descriptors for the caller to use 
 * are pointed to by p_fd_in and p_fd_out.
 * 
 * @param  shell_cmd  the shell command to be executed  
 * @param  p_fd_in    the input pipe file descriptor for the caller to 
 *                    use (value of 0 indicates the shell command doesn't
 *                    use stdin)
 * @param  p_fd_out   the output pipe file descriptor for the caller to 
 *                    use (value of 0 indicates the shell command doesn't
 *                    use stdout) 
 * @return            the parent process id
 */      
pid_t popen2(const string &shell_cmd, int *p_fd_in, int *p_fd_out);

/**
 * Reads the output of a system command.
 *
 * @param s  shell command, will also hold the resulting output
 * @return   true<=> received signal to kill (errno set to EINTR), false otherwise
 */
bool system_output(string &s);
//COME BACK: one version has &cmdReturnVal !!

/**
 * Sets up the server and clients and spawns the corresponding threads.
 * Updates progress flags for the game & taskCount on the computers 
 *
 * @param arg         pointer to the game that is to be played 
 */
void *playGame(void *arg);

/**
 * Makes the ssh call to the server and stores the result of this call 
 * as the server's result for this particular game. (not fully implemented yet)
 * Writes all received results to disk files.
 * 
 * @param arg  pointer to the game that is being played
 */
void *playServer(void *arg);

/**
 * Makes the ssh call to a client and stores the result of this call 
 * as this client's result for this particular game.
 * Writes all received results to files.
 *
 * @param arg  pointer to the game that is being played
 */
void *playClient(void *arg);

/**
 * Updates the contents of the .gam file with inorder game results 
 * as they complete.
 *
 * @param filename        the name of the configuration file with .gam suffix appended
 * @param single          flag indicating one or two games per map
 * @param gameType        the type of game (1, 2, or 3)
 */
void updateGameFile(string filename, bool single, int gameType);

/**
 * Displays the summary of game results in matrix form to stdout
 *
 * @param filename        the base name of the file
 * @param gameType        the type of game (1, 2, or 3)
 */
void displayMatrix(string filename, int gameType);

/**
 * Sets up the alarm for the signal handler
 *
 * @param signo
 */
void sighand(int signo);


/**
 * Returns the next queued game.
 *
 * @param  gameCount      the total number of games in game_queue
 * @return                the index of the next game in game_queue that is queued
 */
int getNextQueuedGame();

/**
 * Checks for available computers.
 * If there are available ones, assign them to new games and mark as unavailable.
 *
 * @param gameType         the type of game (!,2,or 3)
 * @param max__client_jobs the max number of jobs allowed per client machine
 * @param max_server_jobs  the max number of jobs allowed per server machine
 * @return                 Return true if we are done all the games
 */
bool checkComputers(int gameType, int max_client_jobs, int max_server_jobs);

#endif
