#ifndef TRTerrain_H
#define TRTerrain_H

// $Id: TRTerrain.H,v 1.2 2006/12/31 04:09:34 orts_furtak Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

// simple grid-based terrain representation

#include "GameObj.H"
#include "Base.h"
#include "SegmentMerge.H"
#include "TerrainBasicImp.H"
#include <boost/shared_ptr.hpp>
#include "SimpleTerrain.H"

namespace TRTerrain {

  typedef TerrainBase::Loc      Loc;
  typedef TerrainBase::Segment  Segment;  
  typedef TerrainBase::Task     Task;

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
  // fixme: Transfer this functionality into ST_Task, this is a heritage from before
  // the TerrainBase interface.
  struct UnitPath
  {
    const Object *obj;
    Vector<Loc> path;
    Loc goal;
    int current_target; // index of current way point in path vector
    uint4 wakeup; // if SDL_GetTicks() >= wakeup then this UnitPath is not
                  // sleeping ie not to be skipped over.
    int tries; // this is incremented upon unexpected stop and retry
    bool dead; // path has been tried enough times and will be deleted
    Loc spot; // spot of the last unexpected collision

    UnitPath(const Object *obj_, const Vector<Loc> &path_);
    ~UnitPath() {};

    void mark_spot(const Loc &p) { spot = p; };
    void kill() { dead = true; };
    void sleep(int time) { wakeup = SDL_GetTicks() + time; };
    bool sleeping() const { return wakeup > SDL_GetTicks(); };

    //Vector<Segment> intersect(const Segment &);
    bool is_intersect(const Segment &) const;
  };


  /// The internal data complementing Task data:
  struct TR_Task {
    Task task;
    bool isPending;
    bool isFailed;
    boost::shared_ptr<UnitPath> path;    // todo: replace with a cleaner implementation

    TR_Task(const Task& task_) 
      : task(task_), isPending(true), isFailed(false) {};
    TR_Task(const TR_Task& other) 
      : task(other.task), isPending(other.isPending),
      isFailed(other.isFailed), path(other.path) {};
    void init_path(const Object *obj_, const Vector<Loc> &path_) { path=boost::shared_ptr<UnitPath>(new UnitPath(obj_, path_)); };
    //~ST_Task() { delete path; };
    bool isActive() const { return !isPending && !isFailed && path && !path->dead; };
  private:
    TR_Task operator=(const TR_Task& other);
  };

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

//DJD: object's position in the triangulation and its influence {

//keeps track of an object and the triangle it currently occupies and its influence
struct TR_Object
{
	//the ID of the object being monitored
	ObjId id;
	//the triangle in which the object lies
	TraFace *face;
	//the influence of the object
	float influence;

	//constructor - initializes member variables
	TR_Object(ScriptObj *_obj, TraFace *_face = NULL) : id (id = GameObj::get_obj_id(_obj)), face (_face)
	{
		CalculateInfluence(_obj);
	}

	//calculates the object's influence based on its attributes
	void CalculateInfluence(ScriptObj *_obj)
	{
		//get the game object associated with the object
		GameObj *gob = _obj->get_GameObj();
		//get its hit points and weapon
		const sintptr *hp = gob->get_int_ptr("hp");
		ScriptObj *weapon = gob->component("weapon");
		//if it has no weapon, it has no influence
		if (weapon == NULL)
		{
			influence = 0.0f;
			return;
		}
		//get the minimum and maximum damage and cooldown for the weapon
		const sintptr *min = weapon->get_int_ptr("min_damage");
		const sintptr *max = weapon->get_int_ptr("max_damage");
		const sintptr *cooldown = weapon->get_int_ptr("cooldown");
		//if any value failed, it has no influence
		if ((hp == NULL) || (min == NULL) || (max == NULL) || (cooldown == NULL))
		{
			influence = 0.0f;
		}
		else
		{
			//otherwise the influence is calculated as the average damage of its weapon, divided by
			//the number of turns between its attacks, times its hit points
			influence = ((float)(*min) + (float)(*max)) / 2.0f * (float)(*hp) / (float)((*cooldown) + 1);
		}
	}
};

//object's position in the triangulation and its influence }

//DJD: an obstacle's id within ORTS and the triangulation {

//keeps track of the ID of an obstacle within the triangulation so it can be removed
struct TR_Obstacle
{
	//the ID of the obstacle within ORTS
	ObjId ortsId;
	//the ID of the obstacle within the triangulation
	int dcdtId;

	//constructor - initializes the member variables
	TR_Obstacle(ObjId _ortsId, int _dcdtId) : ortsId (_ortsId), dcdtId (_dcdtId) {}

	//returns whether or not the given object is an obstacle
	static bool IsObstacle(GameObj *gob)
	{
		//TODO: should be generalized for use in ORTS outside the tournament
		//should consider resources, buildings, and trees as obstacles
		return ((gob->has_attr("is_mobile")) && (gob->get_int("is_mobile") == 0) &&
			(gob->get_zcat(NULL) == Object::ON_LAND));
	}
};

//an obstacle's id within ORTS and the triangulation }

  /// A Simple implementation of the TerrainBase interface.
  /** This TerrainBase implementation is using a relatively basic Astar algorithm,
      originally written by Sami Wagiaalla and adapted by David Deutscher. 
      Limitations:
      - Only considers a single Obj per Task (ignoring the second and on obj_ids).
      - Assumes all objects are ON_LAND or IN_AIR objects (and that boundaries do not 
        apply to the latter).
  */
  class TR_Terrain : public TerrainBasicImp<TR_Task>
  {
//DJD: private attributes added for triangulation pathfinding {
  private:
	//the triangulator
	TraDcdt *pfDcdt;
	//used to merge segments received
	SegmentMerge sm;
	//objects known to the pathfinder
	Vector<TR_Object> objects;
	//obstacles known to the pathfinder
	Vector<TR_Obstacle> obstacles;
//private attributes added for triangulation pathfinding }

  public:
    TR_Terrain();
    virtual ~TR_Terrain() {};

    /// add command-line options specific to this implementation, using static methods of class Options
    static void add_options(void);

    //---------------------------------------------------------------------
    /// \name Implementation of relevant parts of the TerrainBase interface
    //---------------------------------------------------------------------
    //@{
    virtual void init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_);

    virtual void add_obj(const Object *obj);
    virtual void update_obj(const Object *obj);
    virtual void remove_obj(const Object *obj);
    virtual void add_segments(const Vector<TRTerrain::Segment> &s);

    virtual void cancel_task(const Object *obj);
    virtual void remove_task(TaskId tid);

    virtual void execute_tasks(Vector<MoveCmd> &cmds, Vector<StatusMsg> &msgs);
    virtual bool plan_tasks(uint4 max_time=0);

    virtual real8 find_path(const Loc &l1, const Loc &l2, sint4 radius, Path *path, ConsiderObjects consider);
    virtual real8 find_path(const Object *obj, const Goal &goal, Path *path, ConsiderObjects consider);
    //@}
    //---------------------------------------------------------------------

#if 0
    CWidget *create_widget() const; // implement in libs/gfxclient/src to avoid linking gfx stuff
                                    // in non-gfx clients
#endif    

  private:

//DJD: function prototypes for updating the terrain for obstacles {
	void AddObstacle(GameObj *gob);
	void RemoveObstacle(GameObj *gob);
	void UpdateAbstraction();
//function prototypes updating the terrain for obstacles }

    // prevent copying as we currently don't need it.
    TR_Terrain(const TR_Terrain& other);
    TR_Terrain operator=(const TR_Terrain& other);

    /// The mechanics of world representation and basic A* path-finding
//    class PFEngine; friend class TR_Terrain::PFEngine;

    //---------------------------------------------------------------------
    /// \name Internal accessories
    //---------------------------------------------------------------------
    //@{
//    std::deque<TIter>::iterator get_best_task(void);
    /** Perform PF for a single pending task, and insert results to data structures. 
        Return true iff did some planning. */
    bool plan_pending_task(void);
    /** Perform PF for a single failed task, and insert results to data structures. 
        Return true iff did some planning. */
    bool plan_failed_task(void);
    /// Invalidate existing paths that intersect a given new boundary
    void invalidate_paths(const Vector<Segment> &segments);
    //@}
    //---------------------------------------------------------------------

//    boost::shared_ptr<PFEngine> pfEngine; ///< The internal Path Finding engine
    std::set<const Object*> objsToStop;   ///< Objects for which we should set a stop-moving action
  };

} // end of namespace SimpleTerrain

REGISTER_TYPEOFS_TerrainBasicImp_OneTime(2501, TRTerrain::TR_Task);
REGISTER_TYPEOFS_TerrainBasicImp_ForEach(2505, TRTerrain::TR_Task);

#endif
