// $Id: SoundMixer.C,v 1.6 2006/04/03 21:32:16 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro
// based on David Olofson's simplemixer code

#include "SoundMixer.H"
#include "Options.H"

using namespace std;

REGISTER_TYPEOF(250, Vector<SoundMixer::Voice>::iterator);

Sound::~Sound() {
  if (data) { SDL_FreeWAV(data); data = 0; }
}

uint4 SoundMixer::play(Sound &sound, real4 lvol, real4 rvol)
{
  SDL_LockAudio();

  // remove free slots at the end

  while (!voices.empty()) {
    if (voices.back().position >= voices.back().length)
      voices.pop_back();
    else
      break;
  }

  // search for free slot
  
  uint4 i;
  FOR (i, voices.size()) {
    if (voices[i].position >= voices[i].length) break;
  }

  if (i >= voices.size()) voices.push_back(Voice());
  
  voices[i].data     = reinterpret_cast<sint2*>(sound.data);
  voices[i].length   = sound.length / 2;
  voices[i].position = 0;
  voices[i].vol[0]   = (uint4)(lvol * 256.0);
  voices[i].vol[1]   = (uint4)(rvol * 256.0);

  SDL_UnlockAudio();  
  return i;
}


static void sound_mixer_callback(void *ud, Uint8 *stream, int len)
{
  sint2 *buf = reinterpret_cast<sint2*>(stream);
  SoundMixer &sm = *reinterpret_cast<SoundMixer*>(ud);
  
  // clear the buffer
  memset(buf, 0, len);

  // 2 channels, 2 bytes/sample = 4 bytes/frame
  len /= 4;

  // for each sample...
  FORS (s, 2*len) {
    bool smod2 = s & 1;
    sint4 amp = 0;
    // for each voice...
    FORALL (sm.voices, i) {

      SoundMixer::Voice &v = *i;
      if (!v.data) continue;
      if (v.position >= v.length) { v.data = 0; continue; }

      amp += (v.data[v.position] * v.vol[smod2]) >> 8;
      if (smod2) ++v.position;
    }

    // clipping
    if      (amp > max_sint2) amp = max_sint2;
    else if (amp < min_sint2) amp = min_sint2;

    buf[s] = amp;
  }
}


sint4 SoundMixer::open()
{
  if (is_open) ERR("mixer already opened");

  is_open = true;
  
  SDL_AudioSpec as;

  voices.clear(); 

  if (SDL_InitSubSystem(SDL_INIT_AUDIO) < 0) return -2;

  as.freq     = 44100;
  as.format   = AUDIO_S16SYS;
  as.channels = 2;
  as.samples  = 1024;
  as.callback = sound_mixer_callback;
  as.userdata = this;

  if (SDL_OpenAudio(&as, &audiospec) < 0) return -3;

  if (audiospec.format != AUDIO_S16SYS) return -4;

  SDL_PauseAudio(0);
  return 0;
}


void SoundMixer::close()
{
  if (!is_open) REM("mixer not open");
  is_open = false;
  SDL_PauseAudio(1);
  SDL_LockAudio();
  voices.clear();
  SDL_UnlockAudio();
  SDL_CloseAudio();
}


void SoundMixer::flip_endian(uint1 *data, uint4 length)
{
  assert((length & 1) == 0);
  for (uint4 i=0; i < length; i += 2) swap(data[i], data[i+1]);
}


sint4 SoundMixer::load_sound(const string &filename, Sound &sound)
{
  cout << "loading sound: " << filename << endl;
  
  bool failed = false;
  SDL_AudioSpec spec;

  if (sound.data) SDL_FreeWAV(sound.data);

  sound.data = 0;

  if (SDL_LoadWAV(filename.c_str(), &spec, &sound.data, &sound.length) == 0)
    return -1;
  
  if (spec.freq != 44100)
    errstr << "WARNING: File " << filename
	   << " is not 44.1 kHz. Might sound weird..."
	   << endl;

  if (spec.channels != 1) {
    errstr << "Only mono sounds are supported!" << endl;
    failed = true;
  }

  switch(spec.format) {

  case AUDIO_S16LSB:
  case AUDIO_S16MSB:
    if (spec.format != AUDIO_S16SYS)
      flip_endian(sound.data, sound.length);
    break;

  default:
    fprintf(stderr, "Unsupported sample format!\n");
    failed = true;
    break;
  }

  if (failed) {
    SDL_FreeWAV(sound.data);
    sound.data = 0;
    return -2;
  }
  return 0;
}


static bool sound_test_die = false;

static void sound_test_breakhandler(int /*a*/)
{
  sound_test_die = true;
}


int sound_test(const string &path)
{
  const char *bd = "#*#*#*#*#.*.#..*#.*.#.*.#.*.#.*#";  
  //const char *bd = "#...#...#...#..##...#...#...#.##"; // orig
  const char *cl = "....#..*....#....*..#....*..#.**";
  const char *cb = "#...*..#..*...#.#...*..#..*..*#*"; // orig
  //const char *cb = "#.#.*..#.#*#.##.#.#.*..#.#*..*#*";  
  const char *hh = "..#...#...#...#...#...#...#...#."; // orig
  //const char *hh = "#.#.#.#.#.#.#.#.#.#.#.#.#.#.#.#.";
  const char *tk   = ".#.#.#.*.*.*.......#.#.#.*.*.*..";    

  int res;
  Uint32 step = 0;
  Sint32 timer;
  if (SDL_Init(0) < 0) return -1;

  atexit(SDL_Quit);
  signal(SIGTERM, sound_test_breakhandler);
  signal(SIGINT, sound_test_breakhandler);

  SoundMixer sm;
  
  if (sm.open() < 0) {
    errstr << "Couldn't start mixer!" << endl;
    SDL_Quit();
    return -1;
  }

  res = 0;
  
  Vector<Sound> s(5);

  res |= SoundMixer::load_sound(sanitize_path(path+"/sounds/test/808-bassdrum.wav"), s[0]);
  res |= SoundMixer::load_sound(sanitize_path(path+"/sounds/test/808-clap.wav"),     s[1]);
  res |= SoundMixer::load_sound(sanitize_path(path+"/sounds/test/808-cowbell.wav"),  s[2]);
  res |= SoundMixer::load_sound(sanitize_path(path+"/sounds/test/808-hihat.wav"),    s[3]);
  res |= SoundMixer::load_sound(sanitize_path(path+"/sounds/test/tk21.wav"),         s[4]);  

  if (res) {
    sm.close();
    SDL_Quit();
    errstr << "Couldn't load sounds!" << endl;
    return -1;
  }

  /*
   * Of course, playing this stuff would be
   * much better done in the audio callback,
   * since the timing out here jitters a lot.
   *
   * However, this is just a fun hack to get
   * the thing to make some sound, without
   * slapping in a game or something. :-)
   */

  SDL_Delay(200);
  timer = (Sint32)SDL_GetTicks();

  while (!sound_test_die) {

    if ('#' == bd[step])      sm.play(s[0], 1.0, 0.0);
    else if ('*' == bd[step]) sm.play(s[1], 0.0, 1.0);
    
    if ('#' == cl[step])      sm.play(s[1], 0.6, 0.2);
    else if ('*' == cl[step]) sm.play(s[1], 0.2, 0.6);

    if ('#' == cb[step])      sm.play(s[2], 0.3, 0.2);
    else if ('*' == cb[step]) sm.play(s[2], 0.2, 0.3);

    if ('#' == hh[step])      sm.play(s[3], 0.4, 0.4);

    if ('#' == tk[step])      sm.play(s[4], 0.7, 0.3);
    else if ('*' == cb[step]) sm.play(s[2], 0.3, 0.7);
    
    step = (step + 1) % 32;

    timer += 120;
    while (((Sint32)SDL_GetTicks() - timer) < 0) SDL_Delay(10);
  }

  sm.close();
  SDL_Quit();
  return 0;
}
