// $Id: IO_Buffer.C,v 1.15 2006/04/26 01:38:37 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "IO_Buffer.H"
#include "Compress.H"
#include "ShutDown.H"
#include "Random.H"

using namespace std;

#define LEN_FMT "%07d"    // MAX_LEN must fit!
const sint4 LEN_LEN = 7;  // characters needed for length in message

#define TEST 0 

IO_Buffer::IO_Buffer(TCPsocket so_, sint4 clev_, RandSrc *rand_delay_)
  : so(so_), clev(clev_), rand_delay(rand_delay_)
{
  cout << "clev= " << clev << endl;

  if (rand_delay) {
    cout << "delay!" << endl;
  }
  
  i_len = o_len = 0;
  i_bytes = o_bytes = 0;
  done = false;
  read_good = write_good = true;

  buf_mut = SDL_CreateMutex();
  i_delay_buf_mut = SDL_CreateMutex();  
  o_buf_sem = SDL_CreateSemaphore(0);
  input_chars = new uint1[MAX_LEN];

  read_thread  = SDL_CreateThread(IO_Buffer::read_loop, this);
  write_thread = SDL_CreateThread(IO_Buffer::write_loop, this);

  if (rand_delay) i_delay_thread = SDL_CreateThread(IO_Buffer::i_delay_loop, this);
  else i_delay_thread = 0;
}


IO_Buffer::~IO_Buffer()
{
  REM("~IO_Buffer called");
  finish();
#ifdef OS_MAC
	// needed, at least in OS X, to wake up the write thread
	REM("On OS X waking up semaphore");
  SDL_SemPost(o_buf_sem);
#endif
  SDL_KillThread(write_thread);
  SDL_KillThread(read_thread);
  if (i_delay_thread) SDL_KillThread(i_delay_thread);
  write_thread = read_thread = i_delay_thread = 0;
  delete [] input_chars; input_chars = 0;
  SDL_DestroySemaphore(o_buf_sem);
  SDL_DestroyMutex(buf_mut);
  SDL_DestroyMutex(i_delay_buf_mut);  
  REM("~IO_Buffer done");
}


sint8 IO_Buffer::sent_bytes() {
  sint8 v;
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION      
  if (SDL_mutexP(buf_mut)) ERR("mutex 11");
  v = o_bytes;
  if (SDL_mutexV(buf_mut)) ERR("mutex 12");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION  
  return v;
}


sint8 IO_Buffer::rcvd_bytes() {
  sint8 v;
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION      
  if (SDL_mutexP(buf_mut)) ERR("mutex 13");
  v = i_bytes;
  if (SDL_mutexV(buf_mut)) ERR("mutex 14");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION  
  return v;
}


sint4 IO_Buffer::write_buf_len() 
{
  sint4 v;
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION      
  if (SDL_mutexP(buf_mut)) ERR("mutex 13a");
  v = o_buf.size();
  if (SDL_mutexV(buf_mut)) ERR("mutex 14a");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION  
  return v;
}


sint4 IO_Buffer::read_buf_len()
{
  sint4 v;
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION      
  if (SDL_mutexP(buf_mut)) ERR("mutex 13b");
  v = i_buf.size();
  if (SDL_mutexV(buf_mut)) ERR("mutex 14b");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION  
  return v;

}

// add message at front of output buffer
// and return size of buffer (after adding)

sint4 IO_Buffer::write(const string &s)
{
  sint4 ret;

  if (!write_good) return -1; // error
  
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
  if (SDL_mutexP(buf_mut)) ERR("mutex 7");
  o_buf.push_front(s);
  ret = ++o_len;
  if (SDL_mutexV(buf_mut)) ERR("mutex 8");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION

  if (!rand_delay) {
    SDL_SemPost(o_buf_sem); // signal write thread that data is available
  }
  return ret;
}


// remove last message from input buffer
// and return size of buffer (before removal)

sint4 IO_Buffer::read(string &s)
{
  sint4 ret;
  s.clear();

  if (!read_good) return -1; // error
  
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
  if (SDL_mutexP(buf_mut)) ERR("mutex 9");
  ret = i_len;
  if (ret > 0) {
    s = i_buf.front();
    i_buf.pop_front();
    --i_len;
  }
  if (SDL_mutexV(buf_mut)) ERR("mutex 10");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION

  return ret;
}


// encode msg to be sent:
//
//  compressed (clev > 0)  -> C <original length> compressed message
//  uncompressed           -> U uncompressed message
//
//  return 0 iff OK

sint4 IO_Buffer::encode(string &msg, sint4 clev)
{
  sint4 orig_len = msg.size();
  if (orig_len > MAX_LEN) return 1;

  if (clev > 0) {
    
    Compress compress(clev);

    string cs;
    compress.doit(msg, cs);

    if (cs.size() > MAX_LEN) return 2;
    ostringstream os;
    form(os, "C"LEN_FMT, orig_len);
    msg = os.str() + cs;

    // cout << ">>>>> encode: orig len= " << orig_len << " now= " << msg.size() << endl;
    
  } else {

    msg = 'U' + msg;

  }

  return false;
}


sint4 IO_Buffer::decode(string &msg)
{
  if (msg.size() == 0) return 1;

  if (msg[0] == 'C') {

    Decompress decompress;

    // extract original length
    if (msg.size() <= 1+LEN_LEN) return 2;
    char s[LEN_LEN+1];
    strncpy(s, msg.data()+1, LEN_LEN);
    s[LEN_LEN] = 0; // strncpy does not terminate string!
    
    // all characters must be digits
    FORS (i, LEN_LEN)
      if (s[i] < '0' || s[i] > '9') return 3;

    int orig_len;
    assert(sizeof(int) >= 4);    
    if (sscanf(s, "%d", &orig_len) != 1) return 4;

    msg.erase(0, LEN_LEN+1);
    string ds;
    if (decompress.doit(msg, ds, orig_len)) return 5;

    // cout << ">>>>> decode: orig len= " << orig_len << " was= " << msg.size() << endl;
    
    msg = ds;

  } else if (msg[0] == 'U') {

    msg.erase(0, 1);

  } else return 6;

  return 0;
}



sint4 IO_Buffer::blocking_read(void *buffer, sint4 len)
{
  if (!read_good || !write_good) return -1;
  if (len <= 0) ERR("len <= 0");
  sint4 bytes = 0;

  // cout << "READING " << len << " bytes" << endl;
  
  FOREVER {
    sint4 res=0;
    do {
      res = SDLNet_TCP_Recv(so, (reinterpret_cast<uint1*>(buffer))+bytes, len);
    } while ((res < 0) && (!shut_down));
    if (shut_down)
      return -1;
    else if (res <= 0) {
      cerr << ">>>>>>>> READ: "; perror("read"); cout << flush; cerr << endl;
      //continue;
      return -1;
    }

#if 0
    if (res == 0 && errno != 0) {
      cerr << "?????????? READ: "; perror("read"); cout << flush; cerr << endl;
			continue;
    }
#endif
    
    bytes += res;
    len -= res;
    if (len == 0) return bytes;
    if (len < 0) {
      cerr << "$$$$$$$$$$ READ: somehow got too much?!?" << endl;
      return -1;
    }
  }
}

  
sint4 IO_Buffer::blocking_write(void *buffer, sint4 len)
{
  if (!read_good || !write_good) return -1;  
  if (len <= 0) ERR("len <= 0");

  // cout << "WRITING " << len << " bytes" << endl;
  
  sint4 bytes = 0;

  FOREVER {

    sint4 res = SDLNet_TCP_Send(so, (static_cast<uint1*>(buffer))+bytes, len);

    if (res <= 0) {
      cerr << ">>>>>>>> WRITE: "; perror("write"); cout << flush; cerr << endl;
      return -1;
    }
#if 0
    if (res == 0 && errno != 0) {
      cerr << "?????????? WRITE: "; perror("write"); cout << flush; cerr << endl;
    }
#endif
    
    bytes += res;
    len -= res;
    if (len == 0) return bytes;
  }
}

int IO_Buffer::write_loop(void *data)
{
  IO_Buffer &io = *static_cast<IO_Buffer*>(data);
  REM("write loop started");

  if (!io.rand_delay) {

    // no delay
    
    while (!io.done) {

      // wait for a message
      sint4 sw;
      while ((sw = SDL_SemWait(io.o_buf_sem))) {
        REM("IO Buffer awoken prematurely");
      }

      if (shut_down)
        break;

      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 1");
      // get message from o_buf
      string msg;
      if (io.o_len <= 0) ERR("buffer empty!");
      msg = io.o_buf.back();
      io.o_buf.pop_back();
      --io.o_len;
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      IO_Buffer::encode(msg, io.clev);
    
      uint4 len = msg.size();
      uint4 netlen = SDL_SwapBE32(len);

      // send length of message

      if (io.blocking_write(&netlen, sizeof(netlen)) < 0) {
        REM("socket write error 1");
        if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
          REM2("SDLNet_TCP_Send: ", SDLNet_GetError());
        break;
      }

      // send message
      if (io.blocking_write(const_cast<char*>(msg.data()), len) < 0) {
        REM("socket write error 2");
        if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
          REM2("SDLNet_TCP_Send: ", SDLNet_GetError());
        break;
      }

      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 3");
      // update output statistics
      io.o_bytes += sizeof(len) + len;
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 4");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION
    }

  } else {

    // delay messages

    while (!io.done) {

      bool msg_available = false;
      string msg;
      
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 1");
      // get message from o_buf
      if (io.o_len > 1) {
        msg_available = true;
        msg = io.o_buf.back();
        io.o_buf.pop_back();
        --io.o_len;
      }
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      if (msg_available) {
	
        IO_Buffer::encode(msg, io.clev);
        
        // put message in delay buffer
	
        io.o_delay_buf.push_front(msg);
	io.o_delay_times.push_front(SDL_GetTicks() + (*io.rand_delay)());

      } else SDL_Delay(1); // prevent spinning

      // send messages that are due

      uint4 now = SDL_GetTicks();
	
      while (!shut_down) {

        if (io.o_delay_buf.empty()) break;
        
        if (now <= io.o_delay_times.back()) {
          
          // REM2("send delayed message ", SDL_GetTicks());
          string msg = io.o_delay_buf.back();
          io.o_delay_buf.pop_back();
          io.o_delay_times.pop_back();
	  
          uint4 len = msg.size();
          uint4 netlen = SDL_SwapBE32(len);
	  
          // send length of message
	  
          if (io.blocking_write(&netlen, sizeof(netlen)) < 0) {
            REM("socket write error 1");
            if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
              REM2("SDLNet_TCP_Send: ", SDLNet_GetError());
            break;
          }
	  
          // send message
          if (io.blocking_write(const_cast<char*>(msg.data()), len) < 0) {
            REM("socket write error 2");
            if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
              REM2("SDLNet_TCP_Send: ", SDLNet_GetError());
            break;
          }
	  
          // <<<<<<<<<<<<<<<< START CRITICAL SECTION
          if (SDL_mutexP(io.buf_mut)) ERR("mutex 3");
          // update output statistics
          io.o_bytes += sizeof(len) + len;
          if (SDL_mutexV(io.buf_mut)) ERR("mutex 4");
          // >>>>>>>>>>>>>>>> END CRITICAL SECTION
        }
      }
    }
  }

  io.write_good = false;
  REM("write loop stopped");
  return -1;
}


int IO_Buffer::read_loop(void *data)
{
  IO_Buffer &io = *static_cast<IO_Buffer*>(data);
  REM("read loop started");

  if (!io.rand_delay) {

    // no delay
  
    while ((!io.done) && (!shut_down)) {

      uint4 netlen;

      // read length of message
      if (io.blocking_read(&netlen, sizeof(netlen)) < 0) {
        REM("socket read error 1");
        if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
          REM2("SDLNet_TCP_Recv: ", SDLNet_GetError());
        break;
      }

      uint4 len = SDL_SwapBE32(netlen);

      // cerr << "READ len=" << len << endl;
    
      if (len > IO_Buffer::MAX_LEN) {
	REM("message too long -> kill connection");
	break; // terminate thread
      }

      // read message
      if (io.blocking_read(io.input_chars, len) < 0) {
	REM("socket read error 2");
	if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
	  REM2("SDLNet_TCP_Recv: ", SDLNet_GetError());
	break;
      }

      // REM("READ");

      string msg(reinterpret_cast<char*>(io.input_chars), len);
      sint4 r = IO_Buffer::decode(msg);
      if (r) ERR2("decode error", r);
    
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 5");
      // append message
      io.i_buf.push_back(msg);
      ++io.i_len;
      // update input statistics
      io.i_bytes += sizeof(len) + len;
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 6");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION
    }
  
  } else {

    // delay

    while ((!io.done) && (!shut_down)) {

      uint4 netlen;

      // read length of message
      if (io.blocking_read(&netlen, sizeof(netlen)) < 0) {
        REM("socket read error 1");
        if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
          REM2("SDLNet_TCP_Recv: ", SDLNet_GetError());
	break;
      }

      uint4 len = SDL_SwapBE32(netlen);

      // cerr << "READ len=" << len << endl;
    
      if (len > IO_Buffer::MAX_LEN) {
        REM("message too long -> kill connection");
        break; // terminate thread
      }

      // read message
      if (io.blocking_read(io.input_chars, len) < 0) {
        REM("socket read error 2");
        if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
          REM2("SDLNet_TCP_Recv: ", SDLNet_GetError());
        break;
      }

      // REM("READ");

      string msg(reinterpret_cast<char*>(io.input_chars), len);
      sint4 r = IO_Buffer::decode(msg);
      if (r) ERR2("decode error", r);

      // put message in i_delay_buf

      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.i_delay_buf_mut)) ERR("mutex 5");
      io.i_delay_buf.push_front(msg);
      io.i_delay_times.push_front(SDL_GetTicks() + (*io.rand_delay)());
      if (SDL_mutexV(io.i_delay_buf_mut)) ERR("mutex 6");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION
    }
  }
  
  io.read_good = false;
  REM("read loop stopped");
  return -1;
}


// delivers read messages delayed

int IO_Buffer::i_delay_loop(void *data)
{
  IO_Buffer &io = *static_cast<IO_Buffer*>(data);
  REM("i_delay loop started");

  while (!io.done) {

    uint4 now = SDL_GetTicks();

    FOREVER {
      
      bool msg_to_be_delivered = false;
      string msg;
      
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.i_delay_buf_mut)) ERR("mutex 15");
      msg_to_be_delivered = !(io.i_delay_buf.empty() || now < io.i_delay_times.back());
      if (msg_to_be_delivered) {
        msg = io.i_delay_buf.back();
        io.i_delay_buf.pop_back();
        io.i_delay_times.pop_back();
      }
      if (SDL_mutexV(io.i_delay_buf_mut)) ERR("mutex 16");

      if (!msg_to_be_delivered) { SDL_Delay(1); break; }

      // deliver message
      
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 17");
      // append message
      io.i_buf.push_back(msg);
      ++io.i_len;
      // update input statistics
      io.i_bytes += sizeof(sint4) + msg.size();
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 18");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      //REM2("recv delayed message ", SDL_GetTicks());
    }
  }

  REM("i_delay loop stopped");
  return -1;
}
