// $Id: OSL_RecStream.C,v 1.1.1.1 2004/10/11 09:09:20 orts_mburo Exp $

// (c) Michael Buro, mburo@cs.ualberta.ca, licensed under the GPL


#include "OSL_RecStream.H"

using namespace std;

namespace OSL {

  struct Stream {

    FILE *fd ;
    const uint1 *memptr;
    string name;
    
    stack<sint4> ustack;
    sint4 line_num;
    
    void init()
    {
      /*
	This is a kludge to get preprocessor directives to see
	a '#' as the very first character in the file as if it came
	immediately after a newline...also accounted for by line_num
	starting at zero instead of 1.
      */
      
      put_char('\n') ;
      line_num = 0 ;
    }

    Stream(const uint1 *memptr_, const string &name_) : memptr(memptr_), name(name_)
    {
      fd = 0;
      init();
    }

    Stream(FILE *fd_, const string &name_) : fd(fd_), name(name_)
    {
      memptr = 0;
      init() ;
    }
 
    ~Stream() { if (fd) { fclose(fd); fd = 0; } }
    
    sint4 get_line_num() const { return line_num; }

    string get_name() const { return name; }
    
    sint4 get_char()
    {
      sint4 c ;
      
      /*
	Get a new character - either from the 'undo' stack or
	from the file or from the memory pointer.
	
	Return character or EOF (-1)
      */
      
      if (!ustack.empty()) {

	c = ustack.top(); ustack.pop();

      } else
	
	if (fd != 0) {

	  c = getc(fd);
	  if (c < 0) {
	    c = -1;
	    // test: throw RecStream::ReadErrorException(name);
	    if (ferror(fd)) throw RecStream::ReadErrorException(name);
	  }

	} else

	  if (memptr != 0) {

	    c = static_cast<sint4>(*memptr++);

	    if (c == '\0') {
	      memptr = 0;
	      c = -1;
	    }
	  } else c = -1;
      
      if (c == '\n') ++line_num;
      
      return c ;
    }
    
    void put_char(sint4 c)
    {
      ustack.push(c);
      if (c == '\n') --line_num;
    }
    
  };


  string RecStream::get_name() const {
    if (sstack.empty()) throw EmptyStackException();
    return sstack.top()->get_name();
  }

  
  sint4 RecStream::get_line_num() const {
    if (sstack.empty()) throw EmptyStackException();
    return sstack.top()->get_line_num();
  }

  size_t RecStream::get_stack_size() const { return sstack.size(); }
  

  void RecStream::push(FILE *fd, const string &name) {
    Stream *f = new Stream(fd, name);
    sstack.push(f);
  }
  
  void RecStream::push(const uint1 *memptr, const string &name) {
    Stream *f = new Stream(memptr, name);
    sstack.push(f);
  }

  void RecStream::pop() {
    if (sstack.empty()) throw EmptyStackException();
    delete sstack.top(); sstack.pop();
  }

  sint4 RecStream::get_char()
  {
    if (sstack.empty()) throw EmptyStackException();
    
    sint4 c = sstack.top()->get_char();

    /*
      If we hit the end of the current file, then if this is an
      included file, just pop the stream stack and carry on reading
      from the place we left off.  Otherwise just return EOF
    */

    if (c < 0) {  // EOF

      if (sstack.size() > 1) {
	sstack.pop();
	return get_char();
      }

      return -1;
    }

    return c;
  }

  void RecStream::put_char(sint4 c)
  {
    if (sstack.empty()) throw EmptyStackException();
    sstack.top()->put_char(c);
  }

  sint4 RecStream::peek()
  {
    sint4 c = get_char();
    put_char(c);
    return c;
  }
  
}
