/*
 *  graphHierarchy.h
 *  HOG
 *
 *  Created by Nathan Sturtevant on 1/11/05.
 *  Copyright 2005 Nathan Sturtevant. All rights reserved.
 *
 */

#include "graph.h"
#include "path.h"
#include "glUtil.h"
#include <assert.h>

#ifndef GRAPHHIERARCHY_H
#define GRAPHHIERARCHY_H

#include <vector>
#ifdef OS_MAC
#include <OpenGL/gl.h>
#else
#include <GL/glut.h>
#endif
#include "constants.h"

namespace HOG{
// for now we use this method to allow you to choose the abstraction type
enum tAbstractionType {
	kCliqueAbstraction,
	kNeighborAbstraction
};

class graphHierarchy {
public:
  graphHierarchy(graph *g, tAbstractionType _absMethod = kCliqueAbstraction);
  virtual ~graphHierarchy();
	virtual void verifyHierarchy();
  virtual void draw();
  virtual void rebuild();
	virtual int getRevision() { return 0; }
  void toggleDrawAbstraction(int which);
	
	void setAbstractionType(tAbstractionType _absMethod) { absMethod = _absMethod; }
	
  bool pathable(node *from, node *to);
  bool pathable(unsigned int from, unsigned int to);
	path *getQuickPath(node *from, node *to);
  void getParentHierarchy(node *from, node *to, std::vector<node *> &fromChain, std::vector<node *> &toChain);
  node *getNthParent(node *, int n);
  
  graph* getAbstractGraph(int level) { return abstractions[level]; }
  unsigned int getNumAbstractGraphs() { return abstractions.size(); }
  virtual double h(node *a, node *b) = 0;

  void clearMarkedNodes();
  void clearDisplayLists();
	
  // first step to allowing simple map adjustments...
  virtual void removeNode(node *n);
  void removeEdge(edge *e, unsigned int absLevel);
  // this must be called after removeEdge, but removeNode will take care of itself
  void repairAbstraction();
  node *findNodeParent(node *n);
  edge *findEdgeParent(edge *e, unsigned int absLevel);
	
	virtual int getNumPrimitiveActions() = 0;
	virtual node *getNodeForPrimitiveAction(node *, int) = 0;
protected:
	virtual graph *abstractGraph(graph *g);
	virtual graph *cliqueAbstractGraph(graph *g);
	virtual graph *neighborAbstractGraph(graph *g, int width = 1);
	void addNodesToParent(graph *g, node *n, node *parent, int width);
  void addTunnel(node *n, graph *g, node *newNode);
  void renameNodeInAbstraction(node *which, unsigned int oldID);
  int getChildGroups(node *which);
  void splitNode(node *which, int numGroups);
	void addNodeToRepairQ(node *n);
	void removeNodeFromRepairQ(node *n);
	//void removeDiagonalEdgesAroundNode(node *n);
	
  // helper functions for splitting nodes
	void resetLocationCache(node *n);
	void checkAndCreateParent(node *which);
	void insertNodeIntoHierarchy(node *newNode);
  //void transferNode(node *which, node *oldParent, node *newParent, int whichChild);
	void transferGroup(int group, node *oldParent, node *newParent);
	void abstractUpEdge(unsigned int absLevel, edge *e);
  void extractGroupIntoNewNode(node *parent, int group);
  void mergeGroupIntoNeighbor(node *parent, int group, node *neighbor = 0);
  bool checkNeighborClique(node *child, node *neighbor);
  node *findNeighborCliques(node *parent, int group);
  node *findNeighborCliques(node *);
  int getGroupSize(node *parent, int group);
  node *getNodeInGroup(node *parent, int group);
		
  void cleanMemory();
	
  void drawLevelConnections(node *n);
  virtual recVec getNodeLoc(node *n);
  void drawGraph(graph *g);
		
  std::vector<graph *> abstractions;
  std::vector<GLuint> displayLists;
  // temporary Q for nodes that are modified in the abstraction
  // we can then process them for changes all at once.
  std::vector<node *> modifiedNodeQ;
  unsigned long levelDraw;
	
	tAbstractionType absMethod;
};
}
#endif
