// $Id: GameStateModule.C,v 1.17 2006/08/05 22:20:10 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameObj.H"
#include "TCP_Server.H"
#include "TCP_Client.H"
#include "GUI.H"
#include "ShutDown.H"
#include "Options.H"

using namespace std;

#define DISP_VIEW 0

const sint4 MSLEEP = 10; // ms

#define CMDERR(x) { cout << "ERR: " << x << endl; continue; }

// messages

const sint4 GameStateModule::FROM           = EventFactory::new_who(); 
const sint4 GameStateModule::FINISHED_MSG   = EventFactory::new_what();
const sint4 GameStateModule::STOP_MSG       = EventFactory::new_what();
const sint4 GameStateModule::READ_ERROR_MSG = EventFactory::new_what();
const sint4 GameStateModule::VIEW_MSG       = EventFactory::new_what();


GameStateModule::Options::Options()
{
  ::Options::get("-host", host);
  ::Options::get("-port", port);
  ::Options::get("-dmin", dmin);
  ::Options::get("-dmax", dmax);  
  ::Options::get("-smin", smin);
  ::Options::get("-smax", smax);  
  ::Options::get("-clev", clev);

  ::Options::get("-act", act);
  ::Options::get("-nfog", nfog);

  ::Options::get("-savefile", save_file);
  ::Options::get("-loadfile", load_file);
 
  ::Options::get("-id", id); 
}

void GameStateModule::Options::add()
{
  ::Options o("GameStateModule");
  o.put("-host",  string("localhost"),     "host to connect to");
  o.put("-port",  (sint4)7777, "port to connect to");
  o.put("-dmin",  (sint4)0,    "random network delay (msec) minimum");
  o.put("-dmax",  (sint4)0,    "random network delay (msec) maximum (0:dmin)");
  o.put("-smin",  (sint4)0,    "random client sleep (msec) minimum");
  o.put("-smax",  (sint4)0,    "random client sleep (msec) maximum (0:smin)");
  o.put("-clev",  (sint4)5,    "compression level (0..9)");

  o.put("-act",  "generate actions");
  o.put("-nfog", "no fog of war");

  o.put("-savefile", string(""), "file to save state to");
  o.put("-loadfile", string(""), "file to load state from");

  o.put("-id", string(""), "authentication string sent to server");
}


GameStateModule::GameStateModule(const GameStateModule::Options &opt_) : opt(opt_)
{
  state = ST_IDLE;
  tcp_client = 0;
  game = 0;
  changes = new GameChanges;
  action_changes = new ActionChanges;  
  client_died = false;
  quit = false;
  last_action_msg_sent = 0;
  rand_delay = 0;
  rand_sleep = 0;
  
  if (opt.dmin || opt.dmax) {

    if (opt.dmin < 0 || opt.dmax < 0) ERR("dmin < 0 || dmax < 0");
	
    if (opt.dmax == 0) opt.dmax = opt.dmin;
    if (opt.dmax < opt.dmin) ERR("dmax < dmin");

    opt.dmin /= 2; // symmetric up/down delay
    opt.dmax /= 2;

    if (opt.dmin <= 0) opt.dmin = 1;
    if (opt.dmax <= 0) opt.dmax = 1;    

    rand_delay = new UniRandSrc(1234567, opt.dmin, opt.dmax);
  }

  if (opt.smin || opt.smax) {

    if (opt.smin < 0 || opt.smax < 0) ERR("smin < 0 || smax < 0");
	
    if (opt.smax == 0) opt.smax = opt.smin;
    if (opt.smax < opt.smin) ERR("smax < smin");

    if (opt.smax <= 0) opt.smax = 1;    

    rand_sleep = new UniRandSrc(1234567, opt.smin, opt.smax);
  }
}


GameStateModule::~GameStateModule()
{
  delete rand_delay;
  delete rand_sleep;
  delete game;
  delete changes;
  delete action_changes;
  delete tcp_client;
}


bool GameStateModule::recv_view()
{
  if (client_died) {
    ERR("client already dead - shouldn't get here");
    return false;
  }

  if (!tcp_client->is_good()) { client_died = true; REM("died"); }
  
  if (client_died) {
    quit = true;
    call_handlers(Event(FROM, STOP_MSG));
    return true;
  }

  sint4 ret = tcp_client->read(server_msg);
  if (ret < 0)  {
    REM("read error");
    client_died = true;
    quit = true;
    call_handlers(Event(FROM, READ_ERROR_MSG));
    return true;
  }
  if (ret == 0) return false;

  if (state == ST_RECV_VIEW) {

#if DISP_VIEW
    cout << msg << endl;
    cout << "-----------------------------------" << endl;
#endif

    game->apply_view_diff(server_msg, *changes, *action_changes);
    call_handlers(Event(FROM, VIEW_MSG));
    
    if (game->get_finished()) {
      state = ST_IDLE;
      quit = true;
      call_handlers(Event(FROM, FINISHED_MSG));
    }

    if (rand_sleep) {
      sint4 s = (*rand_sleep)();
      // cout << s << endl;
      SDL_Delay(s);
    }
    return true;
  }
  return false;
}


bool GameStateModule::connect()
{
  if (opt.port <= 1023) ERR("sport <= 1023");

  tcp_client = new TCP_Client(opt.clev, rand_delay);

  if (tcp_client->connect(opt.host.c_str(), opt.port)) {
    ostringstream os;
    os << "can't connect to " << opt.host << ":" << opt.port;
    ERR(os.str());
  }

  client_died = false;
  quit = false;

  // send authentication

  tcp_client->write(opt.id);  
  
  // wait for create message
  
  string msg;

  FOREVER {

    if (tcp_client && client_died) {
      REM("broken connection");
      return false;
    }

    sint4 ret = tcp_client->read(msg);
    if (ret < 0) {
      REM("read error");      
      client_died = true;
      return false;
    }
    if (ret > 0) break;

    SDL_Delay(MSLEEP);
  }

  // message received: cpind + client_world
      
  Game::Options go;

  go.nfog = opt.nfog;
  go.save_file = opt.save_file;
  cout << "SAVE FILE = " << opt.save_file << endl;
  
  stringstream ss(msg);
  ss >> go.cpind;
  if (!ss) ERR("cpind missing");
  
  cout << "CPIND: " << go.cpind << endl;
  
  msg.erase(0, ss.tellg());  // remove cpind from msg
  game = new Game(msg, go);

  state = ST_RECV_VIEW;
  return true;
}


sint4 GameStateModule::send_actions()
{
  if (opt.act) game->gen_client_actions();

  string msg;
  game->gen_action_message(++last_action_msg_sent, msg);
  return tcp_client->write(msg);
}

