/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Global.H"
#include "RT_Server.H"
#include "RT_GameParam.H"
#include "RTS_Game.H"
#include "System.H"
#include "Client.H"

#define SAVE_MSG 0

const sint4 timeout = 10*1000000; // micro seconds

static sint8 start_time = 0;
static sint8 bytes_sent = 0;
static sint8 bytes_rcvd = 0;  
static sint4 frame = 0;

struct IO_Stat {

  sint8 write_t0, write_t1;
  sint4 write_bytes;
  
  sint8 read_t0, read_t1;
  sint4 read_bytes;

};

struct PlayerInfo {

  string name;
  string host;
  sint4  port;
  Client *client;
  Compress compress;      // for objects
  Decompress decompress;  // for actions
  vector<string> strings; // sent/rcvd data
  string end_string;      // end-of-block string
  bool ready;
  bool active;
  uint4 checksum;         // of objs seen by player
  IO_Stat io_stat;
  vector<IO_Stat> io_stat_hist;
  
  PlayerInfo() : compress(8) {
    client = new Client(131072, 131072);
  }

  ~PlayerInfo() { if (client) delete client; }
};


static void write_cstr(const char *s, bool ggs, const string &name, FILE *output)
{
  if (ggs)	{								
    fprintf(output, "tell %s %s\n", name.c_str(), s); fflush(output);	
  } else { 									
    fprintf(output, "%s\n", s); fflush(output);				
  } 
}

static void write_str(const string &s, bool ggs, const string &name, FILE *output)
{
  write_cstr(s.c_str(), ggs, name, output);
}

#define WRITE(x) {                              \
  ostringstream os;				\
  string sss_123;			        \
  os << (x) << flush;			        \
  STR2STRING(os, sss_123);			\
  write_str(sss_123, ggs_mode, from, output);	\
}

#define ERRCONT(x) {				\
  ostringstream os;				\
  string s; 					\
  os << "ERR: " << x;				\
  STR2STRING(os, s);				\
  write_str(s, ggs_mode, from, output);		\
  continue;					\
}

      
static void status(ostream &os)
{
  real4 seconds = (System::real_time()-start_time)/1000000.0;
  form(os,
       "fr=%d @ %.1f, R[%.2f %.1f] W[%.2f %.1f]",
       frame,
       (real8(frame)/seconds),
       bytes_rcvd*0.001/frame, bytes_rcvd*0.001/seconds,
       bytes_sent*0.001/frame, bytes_sent*0.001/seconds);
  os << endl;
}


// true <=> no new active player for usec micro seconds
// false: ready=true in all inactive player infos for which new data arrived
static bool wait_recv(vector<PlayerInfo*> &players, uint4 usec)
{
  fd_set readfds;
  struct timeval timeout;

  timeout.tv_usec = usec % 1000000;
  timeout.tv_sec  = usec / 1000000;
  
  FD_ZERO(&readfds);

  sint4 max_sock = -1;
  uint4 pn = players.size();
  
  FORU (i, pn) {
    PlayerInfo &pinf = *players[i];
    
    if (pinf.client && !pinf.ready && !pinf.active) {
      sint4 socket = pinf.client->socket();
      FD_SET(socket, &readfds);
      max_sock = ::max(max_sock, socket);
    }
  }
  
  if (max_sock < 0) return true;

  sint4 ret = System::select_desc(max_sock+1, readfds,
				  *(fd_set*)0, *(fd_set*)0, &timeout);
  if (ret > 0) {

    FORU (i, pn) {
      PlayerInfo &pinf = *players[i];
      if (pinf.client && !pinf.active) {
	pinf.ready = FD_ISSET(pinf.client->socket(), &readfds);
      }
    }

    return false;
  }

  return true;
}


// read a block of messages from all players (ending with #)

static void recv_msg_block(vector<PlayerInfo*> &players)
{
  sint4 incomplete_n = 0;  // #players which haven't completely sent moves
  sint4 active_n = 0;      // #players with non-empty input buffer
  sint4 reading[players.size()];  // list of incomplete player indices

  FORU (i, players.size()) {
    PlayerInfo &pinf = *players[i];
    if (pinf.client) reading[incomplete_n++] = i;
    pinf.ready = pinf.active = false;
    pinf.strings.clear();
    pinf.io_stat.read_t0 = -1; // to indicate first read time
    pinf.io_stat.read_bytes = 0;
  }
  
#define CLIENT_ERR_CONT(x) {\
  cerr << "player " << pinf.name << " " << x << ", remove" << endl;	\
  delete pinf.client;							\
  pinf.client = 0;							\
  reading[i] = reading[incomplete_n-1];					\
  i--; incomplete_n--;							\
  continue;                                                             \
}

  sint8 starting_time = System::real_time();
  
  while (incomplete_n > 0) {

    // set ready flag for inactive players if select detects that
    // new data arrived

    sint4 ret;
    if (active_n > 0) ret = wait_recv(players, 0);
    else              ret = wait_recv(players, 10000);
    
    if (!ret) {

      // new active connections

      FORS (i, incomplete_n) {
	PlayerInfo &pinf = *players[reading[i]];
	if (pinf.ready &&
	    !pinf.active) {
	  pinf.active = true;
	  if (pinf.io_stat.read_t0 < 0) {
	    pinf.io_stat.read_t0 = System::real_time();
	  }

	  active_n++;
	}
      }
    }

    if (System::real_time() >= starting_time + timeout) {

      FORS (i, incomplete_n) {
	PlayerInfo &pinf = *players[reading[i]];
	cerr << "player " << pinf.name << " read timeout, remove" << endl;
	delete pinf.client;
	pinf.client = 0;
      }
      
      return;
    }
    
    if (active_n > 0) {
	  
      FORS (i, incomplete_n) {

	PlayerInfo &pinf = *players[reading[i]];

	if (pinf.active) {

	  string m;
	  sint4 ret = pinf.client->read_msg2(m, pinf.ready);

	  if (!pinf.ready && ret == 0) { // no data available -> inactive
	    pinf.active = false;
	    active_n--;
	    continue;
	  }
	      
	  pinf.ready = false; // data ready no longer guaranteed

	  if (ret < 0) CLIENT_ERR_CONT("read error");
	  if (ret == 0) continue; // msg not yet completed

	  // got a message!
	      
	  bytes_rcvd += m.size() + 1;
	  pinf.io_stat.read_bytes += m.size() + 1;

	  // assume: (compressed) move messages don't start with '#'!

	  if (m[0] == '#') {  // end of message block

	    pinf.end_string = m;

	    // player done

	    pinf.io_stat.read_t1 = System::real_time();
	    pinf.active = false;
	    active_n--;
	    
	    reading[i] = reading[incomplete_n-1];
	    i--; incomplete_n--;
	    continue;

	  } else {

	    // cout << "received action: <" << m << ">" << endl;
	    pinf.strings.push_back(m);
	  }
	}
      }
    }
  }
}

static void send_msg_block(vector<PlayerInfo*> &players)
{
  FORU (i,  players.size()) {
    PlayerInfo &pinf = *players[i];
    if (pinf.client) {
      uint4 on = pinf.strings.size();
      
      FORU (j, on) {
	pinf.client->write_msg(pinf.strings[j], true);
	bytes_sent += pinf.strings[j].size() + 1;
	
#if SAVE_MSG	
	if (i==0) fprintf(fpsave, "%s ", vis_objs[j].c_str());
#endif	
      }
      pinf.client->write_flush();
    }
  }
}


void server_mode(RTS_Options &opt)
{
  const sint4 LEN=100;
  char line[LEN+1];
  bool running = false;
  vector<PlayerInfo*> players;
  RTS_Game *game = 0;
  RT_GameParam param;
  bool param_ok = false;
  string creator;
  Client *ggs_client = 0;

  string mode;
  opt.get("-m", mode);
  bool ggs_mode = (mode == "ggs");
  FILE *input  = stdin;
  FILE *output = stdout;
  sint4 fdin = 0;
  char CHAR_EOF = '\n';
  
  if (ggs_mode) {
    
    // connect to ggs
    
    ggs_client = new Client(65536, 65536);

    CHAR_EOF = '\\';
    
    string ggsh;
    sint4 ggsp;

    opt.get("-ggsh", ggsh);
    opt.get("-ggsp", ggsp);    
    
    if (ggs_client->connect(ggsh.c_str(), ggsp)) {
      cerr << ggsh << ":" << ggsp << endl;
      ERR("can't connect to GGS");
    }

    fdin = ggs_client->socket();
    input = fdopen(fdin, "r");
    if (!input) ERR("fdopen (read) failed");

    output = fdopen(fdin, "w");
    if (!output) ERR("fdopen (write) failed");
    
    // login

    fprintf(output, "/ants\n"); fflush(output);
    fprintf(output, "blezebub\n"); fflush(output);
  }
  
  FOREVER {
    
    bool blocking = false;  // true <=> waiting for stdin input
    
    FOREVER {
      
      // cerr << "::: running " << running << " blocking " << blocking << endl;

      if (!blocking && running) {
	
	fd_set         fds;
	struct timeval timeout;

	timeout.tv_sec  = 0;
	timeout.tv_usec = 0;
	FD_ZERO(&fds);
	FD_SET(fdin, &fds);
	sint4 ret = System::select_desc(fdin+1, fds, *(fd_set*)0, *(fd_set*)0,
					&timeout, true);
	if (ret <= 0) break;

      } else {

	if (!ggs_mode) cerr << "> " << flush;
      }
      
      if (!fgets(line, LEN, input)) {
	if (feof(input)) { cerr << "input closed" << endl; exit(0); }
      }

      // remove CR+LF in ggs mode, LF in stdin mode
      sint4 l = strlen(line);
      if (l > 0 && line[l-1] == '\n') {
	if (ggs_mode) {
	  if (l == 1 || line[l-2] != '\r')
	    ERR("GGS message corrupt (missing '\r')");
	  line[l-2] = 0;
	} else
	  line[l-1] = 0;
      }

      if (ggs_mode) { // echo
	cerr << "echo: <" << line << ">" << endl;
      }

      istringstream is(line);
      string from, command;

      is >> from;
      if (!is || from.empty()) {
	if (ggs_mode) continue; // ignore messages from central server
	ERRCONT("what? help");
      }

      if (from[from.size()-1] == ':') {

	if (from.size() == 1) continue; // ignore messages from central server
	
	is >> command;
	from.erase(from.size()-1);

      } else {

	if (ggs_mode) continue; // ignore messages from central server
	
	command = from;
	from = "mic";
      }

      cerr << "from= " << from << endl;
      
      blocking = !ggs_mode;

      if (command == "") continue;
      
      if (command == "create") {

	// create new game

	if (param_ok) ERRCONT("game already created");

	ostringstream oserr;
	
	if (param.parse(oserr, is)) {
	  string es;
	  STR2STRING(oserr, es);
	  ERRCONT(es);
	}

	if (RTS_Game::check_param(oserr, param)) {
	  string es;
	  STR2STRING(oserr, es);
	  ERRCONT(es);
	}

	string s;
	ostringstream os;
	os << "game request created: ";
	param.write(os);
	STR2STRING(os, s);
	WRITE(s);
	creator = from;
	param_ok = true;
	running = false;

      } else if (command == "start" ) {

	// start game

	if (from != creator) ERRCONT("only the creator can start a game");
	if (!param_ok) ERRCONT("no request created");
	if (running)    ERRCONT("game already running");
	if (players.size() != uint4(param.n)) ERRCONT("missing player(s)");
	
	opt.put("-type",  param.type); // game type
	opt.put("-x",     param.x);    // width
	opt.put("-y",     param.y);    // height
	opt.put("-pnum",  param.n);    // number of players
	opt.put("-speed", param.s);    // fames/sec

	game = new RTS_Game(opt);
	
	// create games on client side
	
	FORS (i, param.n) {
	  Client &client = *players[i]->client;
	  ostringstream oss;
	  oss << "create " << players[i]->name << " " << i << " ";
	  param.write(oss);
	  string m;
	  STR2STRING(oss, m);
	  client.write_msg(m, true);
	}	  

	frame = 0;
	bytes_sent = 0;
	bytes_rcvd = 0;
	running = true;
	start_time = System::real_time();
	break;
	  
      } else if (command == "stop"  ) {

	// stop game, delete
	  
	if (from != creator) ERRCONT("only creator can stop a game");
	if (!game) ERRCONT("no game");

	delete game;
	game = 0;
	FORALL(vector<PlayerInfo*>, players, i) delete (*i); // closes connections
	players.clear();
	running = false;
	param_ok = false;
	  
      } else if (command == "cont"  ) {

	// continue game
	  
	if (!game)    ERRCONT("no request created");
	if (!running) ERRCONT("game not running");
	break;
	  
      } else if (command == "join"  ) {

	// join game
	  
	if (!param_ok) ERRCONT("no request created");
	if (running)   ERRCONT("game already running");

	string name, host;
	sint4 port;
	
	is >> host >> port;
	if (!is) ERRCONT("syntax error (join host port)");

	if (players.size() >= uint4(param.n)) ERRCONT("no more players");
	
	FORALL (vector<PlayerInfo*>, players, i) if ((*i)->name == from) break;
	if (i != players.end()) ERRCONT("you already joined");

	// connect to "client"
	
	PlayerInfo *pi = new PlayerInfo;

	pi->name = from;
	pi->host = host;
	pi->port = port;

	if (pi->client->connect(host.c_str(), port)) {
	  delete pi;
	  ERRCONT("can't connect");
	}


	// check if host/port is already connected
	{
	  string h1name, h1ip;
	  sint4  h1port;
	  System::getsockhost(h1name, h1ip, h1port, pi->client->socket());

	  FORALL (vector<PlayerInfo*>, players, i) {
	    if ((*i)->port == port) {
	      string h2name, h2ip;
	      sint4  h2port;
	      System::getsockhost(h2name, h2ip, h2port, (*i)->client->socket());
	      if (h1ip == h2ip) break;
	    }
	  }
	  
	  if (i != players.end()) {
	    delete pi;
	    ERRCONT("host/port already in use");
	  }
	}
	
	// OK
	
	players.push_back(pi);
	
      } else if (command == "leave" ) {

	if (!param_ok) ERRCONT("no request created");

	// leave game
	  
	FORALL (vector<PlayerInfo*>, players, i) if ((*i)->name == from) break;
	if (i == players.end()) ERRCONT("you didn't join the game");
	delete (*i);
	players.erase(i);

      } else if (command == "quit"  ) {

	if (from != "mic") ERRCONT("super user command");
	cerr << "bye" << endl;
	exit(0);

      } else if (command == "help") {

	FILE *fp = fopen("help.txt", "r");
	if (!fp) ERRCONT("help file not available");

	string s = "file: help.txt" + CHAR_EOF;

	FOREVER {
	  char line[LEN+1];
	  if (!fgets(line, LEN, fp)) break;
	  sint4 l = strlen(line);
	  if (l > 0 && line[l-1] == '\n') line[l-1] = CHAR_EOF;
	  s += line;
	}

	fclose(fp);

	if (!s.empty()) {
	  if (s[s.size()-1] == CHAR_EOF) s.erase(s.size()-1);
	}
	WRITE(s);
	
      } else if (command == "?") {

	if (!param_ok) {

	  WRITE("no request, no game");

	} else {

	  ostringstream oss;
	  string s;
	  
	  oss << "request created by " << creator << ": ";
	  param.write(oss);
	  oss << CHAR_EOF << players.size() << " player(s): ";
	  FORALL(vector<PlayerInfo*>, players, i) {
	    oss << (*i)->name << " ";
	  }
	  oss << CHAR_EOF << "status: ";

	  if (running) status(oss);
	  else oss << "wait for start";

	  STR2STRING(oss, s);
	  WRITE(s);
	}

      } else {

	ERRCONT("what? -> help");

      }
    }

    if (frame > 0 && (frame % 10) == 0) {
      status(cerr);
    }

    frame++;

    game->compute_visible_elements();

#if SAVE_MSG    
    String f;
    f.form("xxx%d", frame);
    FILE *fpsave = fopen(f.c_str(), "w");
#endif

    // create all views
    
    sint4 winner;
    bool finished = game->finished(winner);
    String end_str;

    sint4 active_n = 0, active_i = -1;
    FORS (i, param.n) {
      if (players[i]->client) { active_n++; active_i = i; }
    }

    if (active_n <= 1) {
      finished = true;
      winner = active_i;
    }

    if (finished) end_str.form("# %d", winner); else end_str = "#";	
    
    FORS (i, param.n) {
      PlayerInfo &pinf = *players[i];
      if (pinf.client) {
	game->create_view_diff(pinf.compress, i, pinf.strings, pinf.checksum);
	pinf.strings.push_back(end_str);
      }
    }

    // send them

    send_msg_block(players);

    game->update_prev_prop();
    
#if SAVE_MSG    
    fclose(fpsave);
#endif

    if (finished) {

      // clean up
      
      bool d;
      opt.get("-disp", d);
      if (d) game->display();
      
      cout << "game ended (";
      if (winner < 0) cout << "draw";
      else cout << "winner is player #" << winner;
      cout << ")" << endl;

      delete game;
      game = 0;
      FORALL(vector<PlayerInfo*>, players, i) delete (*i); // closes connections
      players.clear();
      running = false;
      param_ok = false;
      
    } else {

      game->clear_actions();
      
      recv_msg_block(players);

      FORS (i, param.n) {

	PlayerInfo &pinf = *players[i];

	if (pinf.client) {

#if 0
	  // io statistics
	  real8 time = real8(pinf.io_stat.read_t1-pinf.io_stat.read_t0+10)/1000000.0;
	  
	  form(cerr, "%d: %.2f %d %.2f\n",
	       i,
	       time*1000,
	       pinf.io_stat.read_bytes,
	       real8(pinf.io_stat.read_bytes)*0.001/time
	       );

#endif
	  
	  istringstream is(pinf.end_string.c_str());
	  uint4 cs;
	  string s;
	  is >> s;  // skip #
	  is >> cs;
	  if (!is || cs != pinf.checksum) {
	    cerr << "player " << pinf.name << " cksum error, remove" << endl;
	    delete pinf.client;
	    pinf.client = 0;
	    continue;
	  }
	  game->update_actions(pinf.decompress, i, pinf.strings);
	}
      }
    }

    if (running) {
      
      // execute all actions
      
      bool d;
      opt.get("-disp", d);
      if (d) game->display();
    
      game->exec_actions();
      game->remove_dead();
      game->move_elements();
    }
  }
}
