// $Id: AI_Board.H,v 1.1.1.1 2007/04/09 19:05:43 pjungkun Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#ifndef AI_Board_H
#define AI_Board_H

#include <map>

#include "AI_Object.H"

class Game;
class GameObj;

#define FRIENDLY_SIDE (0)
#define ENEMY_SIDE (1)

#define OTHER_SIDE(X) (!(X))

// different evaluation functions
#define EVAL_UNDEFINED (0)
#define EVAL_NORMAL    (1)
#define EVAL_RANDOM    (2)
#define EVAL_MIX       (3)


/** AI_Board_Area
	Represents information stored in the board about each area
*/
class AI_Board_Area {
public:
  // constructor
  AI_Board_Area();

  void reset();

  // accessors
  bool is_valid() const { return m_valid; }
  bool is_explored() const { return m_explored; }
  bool is_visible() const { return m_visible; }
  bool is_obstacle() const { return m_obstacle; }
  
public:
  // area flags
  bool m_valid;  // if invalid, area data should be ignored
  
  bool m_explored;  // if unexplored, rest of data should be ignored

  bool m_visible;  // note: if visible, then explored...

  bool m_obstacle;  

  // todo: for efficiency, store id's of units in this area

  uint1 h0;  // only used for debugging
  uint1 h1;  // only used for debugging
};


// units on the board, indexed by object ID
typedef std::map<sint4, AI_Object> object_map;

/** The AI_Board class represents our 'board'.
    An AI's abstract view of the world is a board, much like a chess
    board.  A board contains locations (areas) and pieces occupying 
    those areas, as well as important properties of the world.  As
    RTS games might have multiple units very close to each other, it
    is possible that areas may contain multiple units.  This will need
    to be handled, to a degree.

    For now, an 'area' is the same as a 'tile' in the simulation.
    However, this is not necessarily the case and may be changed later.
    Additional implementation would then be required to work with
    more abstract area representations.

*/
class AI_Board {

public:
  // eval option functions
  static std::string get_str_from_eval_opt(sint4 id);
  static sint4 get_eval_opt_from_str(const std::string &s);
  
public:
  AI_Board();

  /** reset all board information to defaults */
  void reset();

  /** update the board to match the current state
      @param game  pointer to object containing current state
  */
  void update_board(Game* game);

  /** retrieves the game pointer associated with the board
  */
  //Game *get_game_ptr() const;

  /** calculate board evaluation */
  double eval(sint4 eval_option) const;

  static const uint4 max_board_size = 64;

  // accessors, using a Point as index
  const AI_Board_Area& get_c(const Point &pt) const;
  AI_Board_Area& get(const Point &pt);

  // accessors, using pair (x, y) as index  
  const AI_Board_Area& get_c(uint4 x, uint4 y) const;
  AI_Board_Area& get(uint4 x, uint4 y);
  
  // move-related
  void move(sint4 obj_id, const Point &pt_to);
  bool is_move_legal(sint4 obj_id, const Point &pt_to) const;
  
  /** check for valid point (bounds check) 
      @return true if invalid, otherwise false
  */
  bool invalid_point(const Point &pt) const;
  bool invalid_point(uint4 x, uint4 y) const;
  
  /** functions with the opposite return values */
  bool valid_point(const Point &pt) const;
  bool valid_point(uint4 x, uint4 y) const;

  /** display board for debugging */
  void print() const;

  /** is the area empty? 
      @return true if empty, true if unknown (fog), otherwise false
      (note that in many cases we are not sure of the correct answer)
  */
  bool area_empty(uint4 x, uint4 y) const;
  bool area_empty(const Point &pt) const;
  
  /** are friendly units grouped up?
      units are defined as grouped if every unit is within
      <n> tiles of all other friendly units.  If <n> is unspecified,
      a "reasonable" default is assumed. (to be determined)
      
      @param n  threshold distance
      
      @return true if grouped up, otherwise false
  */
  bool are_units_grouped(uint4 n = 0);
  
  sint4 count_nearby_enemies(const Point &pt, sint4 r) const;
  sint4 count_nearby_friends(const Point &pt, sint4 r) const;

  uint4 get_width() const { return m_board_width; }
  uint4 get_height() const { return m_board_height; }

  void simulate_random_attacks();
  void simulate_random_motion();
  
  /** add an enemy unit to the fog (in a random location for now) */
  void add_enemies_in_fog();
  
  // is the line of sight blocked between these two areas
  bool line_blocked(const Point &p1, const Point &p2) const;
  
  // total number of enemy objects remaining
  sint4 enemy_count() const { return m_objs[ENEMY_SIDE].size(); }
  
  /** check for victory */
  bool is_game_won() const;
  
  /** check for defeat */
  bool is_game_lost() const;
  
  /** retrieve the flag position
      @param pt    object to receive the return value
      @param side  FRIENDLY_SIDE or ENEMY_SIDE
      
      @return true on success, otherwise false
  */  
  bool get_flag_location(Point &pt, sint4 side) const;

  /** retrieve the post position
      @param pt    object to receive the return value
      @param side  FRIENDLY_SIDE or ENEMY_SIDE
      
      @return true on success, otherwise false
  */  
  bool get_post_location(Point &pt, sint4 side) const;
  
  // allow read access to all AI_Objects
  const object_map &get_friendly_objs() const { return m_objs[FRIENDLY_SIDE]; }
  const object_map &get_enemy_objs() const { return m_objs[ENEMY_SIDE]; }
  const object_map &get_objs(sint4 side) const { return m_objs[side]; }
  
  bool get_nearby_unexplored_point(Point &pt, const Point &pt_start) const;
  bool get_nearby_empty_point(Point &pt, const Point &pt_start) const;

private:
  /** determine obstacles and mark unreachable areas */
  void update_obstacles();

  /** determine if an area is unreachable because of rotation */
  bool unreachable(uint4 x, uint4 y);

  /** marks areas around the point as visible and explored */
  void mark_visible(const Point &pt);

  void fog_entire_map();
  void validate_entire_map();

  void add_obj(const GameObj &gob, bool friendly);
  void get_obj_info(sint4 obj_id, const GameObj &gob, AI_Object &unit);

  void attack_nearby_enemy_of(AI_Object &unit);
  void randomly_move_unit(AI_Object &unit);

  AI_Object* find_obj(sint4 obj_id);
  const AI_Object* find_obj(sint4 obj_id) const;

  void get_random_location_in_fog(Point &pt) const;
  
  double eval_normal() const;
  double eval_random() const;
  
private:
  // for now, use a fixed size array to hold the board areas
  AI_Board_Area m_grid[max_board_size][max_board_size];

  object_map m_objs[2];

  // used as invalid return value
  static AI_Board_Area s_invalid_board_area;

  uint4 m_board_width;
  uint4 m_board_height;

  Game *m_game;  // used to update board, find object ID's, ...

  sint4 m_post_id[2];

  // remember post locations (note: posts can't move)
  Point m_post_loc[2];

  // remember last-known flag location (could be wrong)
  Point m_flag_loc[2];
};

#endif
