// $Id: AI_Client.H,v 1.1.1.1 2007/04/09 19:05:43 pjungkun Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#ifndef AI_Client_H
#define AI_Client_H

/** Description

    AI_Client is a first attempt at creating an ORTS client that is
    able to "play".
    
    As a test game, we will be using the capture-the-flag game, or
    the simplified version that is currently implemented.
*/

#include <map>

#include "AI_Board.H"
#include "AI_Plan.H"

#include "Client.H"
#include "GfxClient.H"

#include "statistics.h"


/** custom ORTS client */
class AI_Client : public GfxClient {

public:
  
  AI_Client(Options &opt);
  virtual ~AI_Client();

  virtual void run_game();
  virtual void on_game_update();

private:
  void load_options();
  void run_without_gui();

  /** generate the AI client's actions to be sent to the server */
  void gen_ai_client_actions();

  /** sets an attack action
      @param gob: our unit
      @param target: (enemy) unit to be attacked
  */
  void attack_target(GameObj &gob, GameObj &target);

  /** sets a pick-up flag action
      @param gob: our post
      @param target: the flag
  */
  void pick_up_flag(GameObj &gob, GameObj &target);

  /** sets a plant flag action
      @param gob: our unit
  */
  void plant_flag(GameObj &gob);
  
  /** sets an attack or move action to chase a unit
      @param gob: our unit
      @param target: (enemy) unit to be chased
  */
  void chase_unit(GameObj &gob, GameObj &target);

  /** moves an object randomly
      @param gob: our unit to be moved
  */
  void move_randomly(GameObj &gob);

  /** restart the planning process */
  void replan();
  
  /** select random plans, find the best of them */
  void find_best_random_plan();
  
  /** select random plans, find the best of them using simulation */
  void find_best_randsim_plan();

  /** create group plans, find the best of them */
  void find_best_group_plan();
  
  /** find best plan: consider some different plans 
      - randomly generated plans not considered
  */
  void find_best_plan();
  void get_best_plan(AI_Plan &plan_out, const AI_Board &bd) const;
  
  /** pick a plan randomly (different than random-plan) */
  void get_randomly_selected_plan(AI_Plan &plan_out, const AI_Board &bd) const;
  
  /** pick a plan randomly for the opponent */
  void get_opponent_plan(AI_Plan &plan_out, const AI_Board &bd) const;
  
  /** pick a *random* plan for the opponent 
      different than above - the plan itself is random
  */
  void get_random_opponent_plan(AI_Plan &plan_out, const AI_Board &bd) const;
  
  /** monte-carlo planning function */
  void find_best_mc_plan();
  
  /** execute the currently active plan 
      can be called repeatedly to continue the execution,
      handle errors in plan execution (possibly repair the plan...)
  */
  void execute_current_plan();

  /** get next waypoint using pathfinding 
    @note searches each call... should save path to save run-time
  */
  Point get_next_waypoint(const Point &ptStart, const Point &ptGoal);

  /** move object to given point */
  void move_to_point(GameObj &gob, const Point &pt);
  
  /** set up the AI_Board to match the current game environment 
      The board's history information is not reset.
  */
  void update_board_from_env();
  
  /** move object to given point */
  void internal_move_to_point(GameObj &gob, const Point &pt);

  /** get object id */
  sint4 get_obj_id(GameObj &gob) const;
  
  /** evaluate a possible plan */
  double evaluate_plan(const AI_Plan &plan, const AI_Board &bd) const;

  /** evaluate a plan using monte-carlo evaluation */
  double evaluate_plan_mc(const AI_Plan &plan, const AI_Board &bd) const;
  
  /** single run of mc evaluation */
  double evaluate_plan_mc_run(const AI_Plan &plan, const AI_Board &bd) const;
  
  /** custom board evaluation function 
      possibly not the same as AI_Board::eval()
      depending on "-eval" option
  */
  double eval_board(const AI_Board &bd) const;
 
  /** return the current time (frame) */
  sint4 get_current_time();
 
  /** return true if plan is complete - else return false */
  bool is_plan_complete(const AI_Plan &plan, const AI_Board &bd) const;

  void simulate_plan_step(
    AI_Plan &plan, 
    AI_Plan &plan_opponent, 
    AI_Board &bd) const;
  
private:
  AI_Board m_board;

  bool m_b_disp_gui;  // whether to display GUI window
  bool m_b_debug;  // whether to print debug info
  bool m_b_timing;  // whether to perform/display timings

  std::map<sint4, AI_Object> m_objects;

  AI_Plan m_plan;  // the currently active plan of action

  // indexed by unit object ID
  std::map<sint4, AI_Move_Command> m_moves;
  
  bool m_b_fighting;  // indicate some fighting is going on

  bool m_b_game_over;  // flag to end the game

  sint4 m_ai_opt;
  sint4 m_eval_opt;
  sint4 m_eval_mode_opt;

  PathFind::Statistics m_stat_run_time;  // for timings

  // number of runs in a monte carlo eval
  // should be set to a large number
  sint4 m_mc_runs;

  // max time per monte carlo run (in simulated time steps)
  sint4 m_mc_steps;
  
  sint4 m_rm_runs;
  sint4 m_rm_dist;
  
  bool m_b_replan_now;  // signals when to replan
  bool m_b_replan_opt;  // whether to replan more often
  
  sint4 m_n_enemies_sighted;  // count enemies currently sighted
                              // used to signal replan
};

#endif
