// $Id: AI_Board.C,v 1.2 2004/10/14 23:43:36 orts_mburo Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL


#include "Global.H"

#include "AI_Board.H"
#include "AI_Common.H"

#include "Game.H"
#include "GameObj.H"
#include "GameTile.H"

using namespace std;

#define MIX_FACTOR (100.0)


//---------------------------------------------------------------------
// Eval option Identifiers
//---------------------------------------------------------------------

static const struct {
  char* eval_name;
  sint4 eval_id;
} eval_opt_map[] = {
  {"normal", EVAL_NORMAL},
  {"random", EVAL_RANDOM},
  {"mix",    EVAL_MIX}
};

string AI_Board::get_str_from_eval_opt(sint4 id)
{
  for (uint4 i = 0; i < sizeof(eval_opt_map) / sizeof(eval_opt_map[0]); i++) {
    if (id == eval_opt_map[i].eval_id) {
      return eval_opt_map[i].eval_name;
    }
  }
  
  return "";
}

sint4 AI_Board::get_eval_opt_from_str(const string &s)
{
  for (uint4 i = 0; i < sizeof(eval_opt_map) / sizeof(eval_opt_map[0]); i++) {
    if (s == eval_opt_map[i].eval_name) {
      return eval_opt_map[i].eval_id;
    }
  }
  
  return EVAL_UNDEFINED;
}


//----------------------------------------------------------------------
// Helper (utility) functions
//----------------------------------------------------------------------

/** calculate a percent chance
    @param p   % chance between 0 and 100
*/

#if 0
static bool percent_chance(uint4 p) {
  return (Rand.rand_uint4() % 100) < p;
}
#endif

// approx. distance check, r in area-units
static bool within_distance(const Point &p1, const Point &p2, sint4 r)
{
  double dx = p2.x - p1.x;
  double dy = p2.y - p1.y;
  
  double d_sq = (dx * dx) + (dy * dy);
  double r_sq = r * r;
  
  return d_sq <= r_sq;
}


//----------------------------------------------------------------------
// AI_Board_Area implementation
//----------------------------------------------------------------------

AI_Board_Area::AI_Board_Area()
{
  reset();
}

// reset to default (initial) state
void AI_Board_Area::reset()
{
  m_valid = false;
  
  m_explored = false;
  m_visible = false;
  
  m_obstacle = false;
}


//----------------------------------------------------------------------
// AI_Board implementation
//----------------------------------------------------------------------

// used as invalid return value
AI_Board_Area AI_Board::s_invalid_board_area;

AI_Board::AI_Board()
{
  reset();
}

void AI_Board::reset()
{
  // reset entire board to default values
  FORU (i, max_board_size) {
	FORU (j, max_board_size) {
	  m_grid[i][j].reset();
	}
  }

  m_board_width = 0;
  m_board_height = 0;
  
  m_game = NULL;
  
  FORU(i, 2) {
    m_objs[i].clear();
    
    m_post_id[i] = -1;
    
    memset(&m_post_loc[i], -1, sizeof(Point));
    memset(&m_flag_loc[i], -1, sizeof(Point));
  }
}

const AI_Board_Area& AI_Board::get_c(const Point &pt) const
{ 
  return get_c(pt.x, pt.y);
}

AI_Board_Area& AI_Board::get(const Point &pt)
{ 
  return get(pt.x, pt.y);
}

const AI_Board_Area& AI_Board::get_c(uint4 x, uint4 y) const
{
  if (invalid_point(x, y)) {
	return s_invalid_board_area;
  }

  // return a reference
  return m_grid[y][x];
}

AI_Board_Area& AI_Board::get(uint4 x, uint4 y)
{
  if (invalid_point(x, y)) {
	return s_invalid_board_area;
  }

  // return a reference
  return m_grid[y][x];
}

void AI_Board::update_board(Game* game)
{
  // incremental update function
  m_game = game;
  
  if ((0 == m_board_width) || (0 == m_board_height)) {
    m_board_width = game->get_map().get_width();
	m_board_height = game->get_map().get_height();
	
	// enable the 'valid' flags
	validate_entire_map();
  }

  // update map visibility - start by fogging the entire map
  fog_entire_map();
  
  // clear objects: todo: remember old object positions...
  m_objs[FRIENDLY_SIDE].clear();
  m_objs[ENEMY_SIDE].clear();
  
  sint4 p = game->get_client_player();  // our player index
  sint4 player_n = game->get_player_num();

  // loop through all objects for all players
  FORS(q, player_n) {

    const Game::ObjCont &obj = game->get_objs(q);
    
    FORU (i, obj.size()) {
      
      GameObj &gob = *obj[i];
      if (TYPE_GLOBAL == get_obj_type(gob))
        continue;  // skip global objects
      
	  // add units/objects and clear fog where required
	  
      if (p == q) {
        // friendly
		add_obj(gob, true /*friendly*/);
		
		Point pt_board = get_board_position(gob);
        mark_visible(pt_board);

	  } else {
        // enemy
        add_obj(gob, false /*enemy*/);
      }
      
    }  // end loop through objects
    
  } // endloop through players

  update_obstacles();
}

double AI_Board::eval(sint4 eval_opt) const
{
  switch (eval_opt) {
    case EVAL_NORMAL:
      return eval_normal();
    
    case EVAL_RANDOM:
      return eval_random();
    
    case EVAL_MIX:
      return eval_normal() + (MIX_FACTOR * eval_random());
    
    default:
      ERR("undefined eval_opt");
      break;
  }
  
  return 0.0;
}

double AI_Board::eval_normal() const
{
  // Recent changes:
  //   add value for unit being alive
  //   reduce 'flag factors' values significantly
  
  // constants, factors, importance of features...
  const double k_explored_bonus = 0.001;
  const double k_visible_bonus = 0.0001;
  
  const double k_flag_post_dist_factor = 0.02;
  
  // factor to encourage at least one unit to approach enemy flag
  const double k_flag_unit_dist_factor = 0.01;
  
  // factor to encourage possession of enemy flag
  const double k_flag_value = 0.1;
  
  // value of a unit being alive
  //const double k_unit_alive_value = 1.0;
  
  double val = 0;
  
  // give bonuses for explored/visible areas
  FORU (y, m_board_height) {
    FORU (x, m_board_width) {
      
	  const AI_Board_Area &area = get_c(x, y);
	  
	  if (area.m_explored) {
		val += k_explored_bonus;  // give a small bonus
	  }
	  
	  if ((!area.m_obstacle) && (area.m_visible)) {
		val += k_visible_bonus;  // give an additional bonus - keeping an eye...
	  }
	  
	} // FORU
  } // FORU
  
  double min_dist_to_enemy_flag = MAX_DIST;
  double min_dist_to_friendly_flag = MAX_DIST;
  
  // loop through objects
  object_map::const_iterator obj_iter = m_objs[FRIENDLY_SIDE].begin();
  for (; obj_iter != m_objs[FRIENDLY_SIDE].end(); obj_iter++) {

	// friendly unit
	const AI_Object &obj = obj_iter->second;
	
    switch (obj.m_obj_type) {
    case TYPE_UNIT:
      //val += k_unit_alive_value + (0.01 * obj.m_hp);  // friendly material
      //val += 1.1 * sqrt(0.01 * obj.m_hp);  // friendly material
      val += sqrt(0.01 * obj.m_hp);  // friendly material
	  
      // it's good that we have at least one enemy nearby to kill
      //if (count_nearby_enemies(obj.location(), ATTACK_RANGE_TILES) > 0) {
      //  val += 0.21;
      //}
      
      if (obj.m_has_flag) {
        // we possess the enemy flag!
        val += k_flag_value;
        min_dist_to_enemy_flag = 0;
        
        if (!is_equal(m_flag_loc[ENEMY_SIDE], obj.location())) {
          DB(obj.hp());
          ERR("!");
        }
      }
      else {
        double dist = calc_distance(obj.location(), m_flag_loc[ENEMY_SIDE]);
        if (dist < min_dist_to_enemy_flag) {
          min_dist_to_enemy_flag = dist;
        }
      }
      break;
    
    case TYPE_POST:
      if (!is_equal(m_post_loc[FRIENDLY_SIDE], obj.location())) ERR("!");
      break;
    
    case TYPE_FLAG:
      if (obj.has_flag() && !is_equal(m_flag_loc[FRIENDLY_SIDE], obj.location())) ERR("!");
      break;
    
    default:
      ERR("unhandled case");
      break;
    }
  }
  
  object_map::const_iterator enemy_iter = m_objs[ENEMY_SIDE].begin();
  for (; enemy_iter != m_objs[ENEMY_SIDE].end(); enemy_iter++) {
    
	// enemy unit
	const AI_Object &obj = enemy_iter->second;
    switch (obj.m_obj_type) {
    case TYPE_UNIT:
      val -= sqrt(0.01 * obj.m_hp);  // enemy material
        
      // it's bad if this enemy is close enough to shoot at us
      //if (count_nearby_friends(obj.location(), ATTACK_RANGE_TILES) > 0) {
        // note that 1v1 battles are encouraged rather than always
        // running away
        //val -= 0.2;
      //}
      
      if (obj.m_has_flag) {
        // this enemy unit has our flag!
        val -= k_flag_value;
        min_dist_to_friendly_flag = 0;
        if (!is_equal(m_flag_loc[FRIENDLY_SIDE], obj.location())) ERR("!");
      } else {
        double dist = calc_distance(obj.location(), m_flag_loc[FRIENDLY_SIDE]);
        if (dist < min_dist_to_friendly_flag) {
          min_dist_to_friendly_flag = dist;
        }
      }
      break;

    case TYPE_POST:
      if (!is_equal(m_post_loc[ENEMY_SIDE], obj.location())) ERR("!");
      break;
    
    case TYPE_FLAG:
      if (obj.has_flag() && !is_equal(m_flag_loc[ENEMY_SIDE], obj.location())) ERR("!");
      break;
      
    default:
      DB(obj.m_obj_type);
      ERR("unhandled case");
      break;
    }
	
  } // enemy unit loop
  
  // sanity checks
  if (m_post_loc[FRIENDLY_SIDE].x < 0) {
    ERR("could not locate friendly post");
  }
  
  if (m_post_loc[ENEMY_SIDE].x < 0) { 
    ERR("could not locate enemy post");  // expected at beginning
    // enemy post loc should be known
  }
  
  // add a term for the distance between flag and enemy post
  // the lower the distance, the closer the flag is to being captured...
  
  //DB(val);
  
  if (valid_point(m_post_loc[FRIENDLY_SIDE]) && 
      valid_point(m_flag_loc[FRIENDLY_SIDE]) &&
      valid_point(m_post_loc[ENEMY_SIDE]) &&
      valid_point(m_flag_loc[ENEMY_SIDE])) {
        
    double d1 = calc_distance(m_post_loc[FRIENDLY_SIDE], m_flag_loc[ENEMY_SIDE]);
    double d2 = calc_distance(m_post_loc[ENEMY_SIDE], m_flag_loc[FRIENDLY_SIDE]);
    double diff1 = d2 - d1;
    val += k_flag_post_dist_factor * (diff1);
     
    double diff2 = min_dist_to_friendly_flag - min_dist_to_enemy_flag;
    val += k_flag_unit_dist_factor * (diff2);
  }
  
  //DB(val);
  //cout << endl;
  
  return val;
}

double AI_Board::eval_random() const
{
  double val = (double)Rand.rand_real8();
  
  // debugging...
  //cout << "random val: " << val << endl;
  
  return val;
}

void AI_Board::move(sint4 obj_id, const Point &pt_to)
{
  if (invalid_point(pt_to)) {
    DB_PT(pt_to);
    ERR("invalid move");
    return;
  }

  AI_Object* p_unit = find_obj(obj_id);
  if (NULL == p_unit) {
	ERR("invalid object ID");
	return;
  }
  
  p_unit->m_location = pt_to;
  
  if (p_unit->is_friendly()) {
	// update visibility flags
	mark_visible(pt_to);
    
    //cout << "marking visibility" << endl;
    
    if (p_unit->has_flag()) {
      m_flag_loc[ENEMY_SIDE] = pt_to;
    }
  } else {
    if (p_unit->has_flag()) {
      m_flag_loc[FRIENDLY_SIDE] = pt_to;
    }  
  }
}

bool AI_Board::is_move_legal(sint4 obj_id, const Point &pt_to) const
{
  const AI_Object *p_unit = find_obj(obj_id);
  if (NULL == p_unit)
	return false;  // invalid object ID
  
  if (invalid_point(pt_to))
    return false;
  
  const AI_Board_Area &area = get_c(pt_to);
  if (!area.is_valid())
	return false;
  
  if (area.is_obstacle())
	return false;
  
  return true;
}

bool AI_Board::invalid_point(const Point &pt) const
{
  if ((pt.x < 0) || (pt.y < 0))
    return true;

  return invalid_point(pt.x, pt.y);
}

bool AI_Board::invalid_point(uint4 x, uint4 y) const
{
  return (x >= m_board_width) || (y >= m_board_height);
}

bool AI_Board::valid_point(const Point &pt) const
{
  if ((pt.x < 0) || (pt.y < 0))
    return false;

  return valid_point(pt.x, pt.y);
}

bool AI_Board::valid_point(uint4 x, uint4 y) const
{
  return (x < m_board_width) && (y < m_board_height);
}

void AI_Board::print() const
{
  cout << "current board [" << m_board_width << "x" << m_board_height << "]";
  cout << endl;

  char units[max_board_size][max_board_size];
  memset (units, 0, sizeof(units));
  
  object_map::const_iterator iter = m_objs[FRIENDLY_SIDE].begin();
  for (; iter != m_objs[FRIENDLY_SIDE].end(); iter++) {
	const AI_Object &unit = iter->second;
	uint4 x = unit.m_location.x;
	uint4 y = unit.m_location.y;
	
	if (0 == units[y][x]) {
	  units[y][x] = 'x';
	} else {
	  units[y][x] = '&';
	}
  }
  
  iter = m_objs[ENEMY_SIDE].begin();
  for (; iter != m_objs[ENEMY_SIDE].end(); iter++) {
	const AI_Object &unit = iter->second;
	uint4 x = unit.m_location.x;
	uint4 y = unit.m_location.y;
	
	if (0 == units[y][x]) {
	  units[y][x] = 'o';
	} else {
	  units[y][x] = '&';
	}
  }
  
  FORU (y, m_board_height) {
    
    FORU (x, m_board_width) {

      const AI_Board_Area &area = get_c(x, y);
	  
	  //cout << "|" << is_zero(area.h0) << " " << is_zero(area.h1);
	  
	  if (!area.is_valid()) {
		cout << " ! ";
	  } else if (units[y][x] != 0) {
		cout << " " << units[y][x] << " ";
	  } else if (!area.is_explored()) {
		cout << " ? ";
	  } else if (area.is_obstacle()) {
		cout << "|||";
	  } else {
		cout << " . ";
	  }
      
    }
    
    cout << endl;
  }
  
  cout << "." << endl;
}

bool AI_Board::area_empty(uint4 x, uint4 y) const
{
  return area_empty(Point(x, y));
}

bool AI_Board::area_empty(const Point &pt) const
{
  if ((pt.x < 0) || (pt.y < 0))
    return false;
  
  const AI_Board_Area &area = get_c(pt);
  if (!area.is_valid())
	return false;

  if (area.m_obstacle)
	return false;
  
  // check for units occupying this area
  object_map::const_iterator iter = m_objs[FRIENDLY_SIDE].begin();
  for (; iter != m_objs[FRIENDLY_SIDE].end(); iter++) {
	if (iter->second.m_location == pt)
	  return false;
  }
  
  iter = m_objs[ENEMY_SIDE].begin();
  for (; iter != m_objs[ENEMY_SIDE].end(); iter++) {
	if (iter->second.m_location == pt)
	  return false;
  }
  
  return true;
}

void AI_Board::update_obstacles()
{
  const Map<GameTile> &map = m_game->get_map();

  //GameTile *pTiles = m_game->get_tiles();
  //if (NULL == pTiles) ERR("unexpected null return value");

  // loop through each tile
  FORU (y, m_board_height) {
	FORU (x, m_board_width) {

	  AI_Board_Area &area = get(x, y);
	  
	  // take rotation into account - some areas are not reachable
	  // (consider them obstacles)
	  if (unreachable(x, y)) {
		area.m_explored = true;
		area.m_obstacle = true;
	  }
		
      const GameTile& tile = map(x, y);

      if (!area.is_explored()) {
        //cout << "area unexplored, ";
        //DB((int)tile.type1);
        //DB(tile.h1);
        //DB(tile.h2);
        //cout << endl;
		continue;
      }
      
	  // check for unreachable areas
      if ((0 == tile.get_typeW()) || (0 != tile.get_min_h()) || (0 != tile.get_max_h())) {
        // in this case, the tile is considered an obstacle...
      
		// flag this area as an obstacle
		area.m_obstacle = true;
		
		/*
		// (optional and experimental) flag adjacent areas as obstacles
		get(x - 1, y).m_obstacle = true;
		get(x + 1, y).m_obstacle = true;
		get(x, y - 1).m_obstacle = true;
		get(x, y + 1).m_obstacle = true;
		*/
      }

    }  // FORU
  }  // FORU

}

bool AI_Board::unreachable(uint4 x, uint4 y)
{
  // possible off-by-one error - only approximate
  
  if (y < (0.5 * m_board_height) - x)
	return true;  // top left corner
  
  if (y >= (0.5 * m_board_height) + x)
	return true;  // bottom left corner
  
  if (y <= (-0.5 * m_board_height) + x)
	return true;  // top right corner
  
  if (y + 1 >= (1.5 * m_board_height) - x)
	return true;  // bottom right corner
  
  return false;
}
	
void AI_Board::mark_visible(const Point &pt)
{
  // point pt is currently occupied by a friendly
  
  // mark nearby areas as visible and explored (however, they could be
  // blocked from view, so try to handle most of those cases)
  const sint4 max_d = 3;
  
  for (sint4 dx = -max_d; dx <= max_d; dx++) {
    for (sint4 dy = -max_d; dy <= max_d; dy++) {
	  
	  Point pt_curr(pt.x + dx, pt.y + dy);
	  
	  if (line_blocked(pt, pt_curr))
		continue; // not visible - line of sight is blocked
	  
      AI_Board_Area &area = get(pt_curr);
	  area.m_visible = true;
	  
	  // if visible, then explored must be true as well
	  area.m_explored = true;
    }
  }
}

sint4 AI_Board::count_nearby_enemies(const Point &pt, sint4 r) const
{
  sint4 count = 0;

  // loop through enemy units
  object_map::const_iterator iter = m_objs[ENEMY_SIDE].begin();
  for (; iter != m_objs[ENEMY_SIDE].end(); iter++) {
	
	const AI_Object &curr_unit = iter->second;
    if (TYPE_UNIT != curr_unit.m_obj_type)
      continue;  // only count units
	
	if (!within_distance(curr_unit.m_location, pt, r))
	  continue;  // too far away
	
	count++;
  }

  return count;
}

sint4 AI_Board::count_nearby_friends(const Point &pt, sint4 r) const
{
  sint4 count = 0;

  // loop through friendly units
  object_map::const_iterator iter = m_objs[FRIENDLY_SIDE].begin();
  for (; iter != m_objs[FRIENDLY_SIDE].end(); iter++) {
	
	const AI_Object &curr_unit = iter->second;
    if (TYPE_UNIT != curr_unit.m_obj_type)
      continue;  // only count units
	
	if (!within_distance(curr_unit.m_location, pt, r))
	  continue;  // too far away
	
	count++;
  }

  return count;
}

bool AI_Board::are_units_grouped(uint4 /*n*/)
{
  // todo: implement this function
  
  return false;
}

void AI_Board::validate_entire_map()
{
  // cover the map in fog
  FORU (i, m_board_height) {
	FORU (j, m_board_width) {
	  m_grid[i][j].m_valid = true;
	}
  }
}

void AI_Board::fog_entire_map()
{
  // cover the map in fog
  
  FORU (i, m_board_height) {
	FORU (j, m_board_width) {
	  m_grid[i][j].m_visible = false;
	}
  }
}

void AI_Board::add_obj(const GameObj &gob, bool friendly)
{
  sint4 obj_id = get_obj_id(gob, m_game);
  
  AI_Object ai_obj;
  get_obj_info(obj_id, gob, ai_obj);
  
  if (ai_obj.m_obj_type <= 0) ERR ("invalid obj type");
  
  // update post and flag info
  if (TYPE_POST == ai_obj.m_obj_type) {
    if (friendly) {
      m_post_id[FRIENDLY_SIDE] = obj_id;
      m_post_loc[FRIENDLY_SIDE] = ai_obj.location();

      //DB(m_friendly_post_id);
      
      // automatically know location of enemy post
      if (is_equal(m_post_loc[FRIENDLY_SIDE], POST_1_LOC)) {
        m_post_loc[ENEMY_SIDE] = POST_2_LOC;
      } else if (is_equal(m_post_loc[FRIENDLY_SIDE], POST_2_LOC)) {
        m_post_loc[ENEMY_SIDE] = POST_1_LOC;
      } else {
        ERR("unexpected location");
      }
      
      // know location of enemy flag (beside post)
      if (m_flag_loc[ENEMY_SIDE].x < 0) {
        m_flag_loc[ENEMY_SIDE] = m_post_loc[ENEMY_SIDE];
        m_flag_loc[ENEMY_SIDE].x++;
      }
      
    } else {
      m_post_id[ENEMY_SIDE] = obj_id;
      
      // check hardcoded post location
      if (!is_equal(m_post_loc[ENEMY_SIDE], ai_obj.location())) {
        ERR("unexpected location");
      }
    }
  }

  if (ai_obj.has_flag()) {
    
    if (TYPE_UNIT == ai_obj.type()) {
      
      if (friendly) {
        m_flag_loc[ENEMY_SIDE] = ai_obj.location();
      } else {
        m_flag_loc[FRIENDLY_SIDE] = ai_obj.location();
      }
      
    } else if (TYPE_FLAG == ai_obj.type()) {
    
      if (friendly) {
        m_flag_loc[FRIENDLY_SIDE] = ai_obj.location();
      } else {
        m_flag_loc[ENEMY_SIDE] = ai_obj.location();
      }
    }
    
  }
  
  if (ai_obj.m_hp <= 0) {
    cout << "warning: object is dead - hp <= 0" << endl;
    ERR("unhandled case");
    return;
  }
  
  ai_obj.m_friendly = friendly;
  
  if (friendly) {
	m_objs[FRIENDLY_SIDE][obj_id] = ai_obj;
  } else {
	m_objs[ENEMY_SIDE][obj_id] = ai_obj;
  }
}

void AI_Board::get_obj_info(sint4 obj_id, const GameObj &gob, AI_Object &ai_obj)
{
  if (TYPE_GLOBAL == ai_obj.m_obj_type) ERR("!");
    
  // make a copy of all information
  ai_obj.m_obj_id = obj_id;
  ai_obj.m_obj_type = get_obj_type(gob);
  
  ai_obj.m_location = get_board_position(gob);
  ai_obj.m_hp = get_hp(gob);
  ai_obj.m_has_flag = has_flag(gob);
}

void AI_Board::simulate_random_attacks()
{
  object_map::iterator iter = m_objs[FRIENDLY_SIDE].begin();
  for (; iter != m_objs[FRIENDLY_SIDE].end(); iter++) {
    attack_nearby_enemy_of(iter->second);
  }
  
  iter = m_objs[ENEMY_SIDE].begin();
  for (; iter != m_objs[ENEMY_SIDE].end(); iter++) {
	attack_nearby_enemy_of(iter->second);
  }
}

void AI_Board::attack_nearby_enemy_of(AI_Object &unit)
{
  const sint4 r = ATTACK_RANGE_TILES;
  const sint4 damage = ATTACK_DAMAGE;  // fixed for now
 
  if (TYPE_FLAG == unit.m_obj_type)
    return;  // flags do not attack
  
  if (unit.m_hp <= 0)
	return;  // we cannot attack, we are dead/dying
  
  object_map::iterator iter;
  object_map::iterator it_end;
  
  if (unit.m_friendly) {
	// friend
	iter = m_objs[ENEMY_SIDE].begin();
	it_end = m_objs[ENEMY_SIDE].end();
  } else {
	// enemy
	iter = m_objs[FRIENDLY_SIDE].begin();
	it_end = m_objs[FRIENDLY_SIDE].end();
  }
  
  for (; iter != it_end; iter++) {
	AI_Object &target_unit = iter->second;
	
	if (unit.m_friendly == target_unit.m_friendly)
	  continue;  // not an enemy
	
	if (!within_distance(unit.m_location, target_unit.m_location, r))
	  continue;  // too far away
	
	if (target_unit.m_hp <= 0)
	  continue;  // already dead/dying
	
	// attack unit
	target_unit.m_hp -= damage;
	
	if (target_unit.m_hp < 0)
	  target_unit.m_hp = 0;
	
    if ((target_unit.m_hp <= 0) && (TYPE_FLAG == target_unit.m_obj_type)) {
      if (target_unit.m_has_flag) {
        // capture the flag
        target_unit.m_has_flag = false;
        unit.m_has_flag = true;
        
        // update flag location
        if (unit.is_friendly()) {
          m_flag_loc[ENEMY_SIDE] = unit.location();
        } else {
          m_flag_loc[FRIENDLY_SIDE] = unit.location();
        }
      }
    }
    
	// break - only attack one unit per call
	break;
  }

}

void AI_Board::simulate_random_motion()
{
  // take each unit and move it randomly to an adjacent tile (or
  // remain still, with equal probability)
  
  object_map::iterator iter = m_objs[FRIENDLY_SIDE].begin();
  for (; iter != m_objs[FRIENDLY_SIDE].end(); iter++) {
	randomly_move_unit(iter->second);
  }
  for (iter = m_objs[ENEMY_SIDE].begin(); iter != m_objs[ENEMY_SIDE].end(); iter++) {
	randomly_move_unit(iter->second);
  }
}

void AI_Board::add_enemies_in_fog()
{
  // find next id (use max + 1)
  sint4 max_id = 0;
  object_map::const_iterator iter = m_objs[ENEMY_SIDE].begin();
  for (; iter != m_objs[ENEMY_SIDE].end(); iter++) {
	if (iter->second.id() > max_id) {
	  max_id = iter->second.id();
	}
  }
  
  sint4 obj_id = max_id + 1;
  
  sint4 total_enemy = 7;
  // todo: this is kind of wrong and not working with flag/post added
  // todo: subtract enemies killed (dead should be removed from board)

  sint4 n_hidden_enemies = total_enemy - enemy_count();
  FORS (i, n_hidden_enemies) {
	
	// add a new enemy unit (with guessed stats/location)
	AI_Object &unit = m_objs[ENEMY_SIDE][obj_id];
	unit.m_obj_id = obj_id;
	unit.m_obj_type = TYPE_UNIT;
    
	unit.m_friendly = false;
	
	get_random_location_in_fog(unit.m_location);
	if (unit.m_location.x < 0)
	  return;  // error finding random location in fog
  
	unit.m_hp = MAX_HP;
  
	unit.m_visible = false;
	
	//cout << "enemy unit added in fog" << endl;
	//DB_PT(unit.m_location);
	
	obj_id++;
  }
}

bool AI_Board::line_blocked(const Point &p1, const Point &p2) const
{
  // we don't need to be perfect, any approximate solution is fine
  
  // right now, only handle up to dist=2
  // and so only need to check one tile (in between the 2)
  
  Point pt;
  pt.x = (p1.x + p2.x) / (Scalar)2;
  pt.y = (p1.y + p2.y) / (Scalar)2;
  
  if (is_equal(pt, p1))
	return false;
  
  if (is_equal(pt, p2))
	return false;
  
  if (get_c(pt).is_obstacle())
	return true;
  
  return false;
}

void AI_Board::randomly_move_unit(AI_Object &unit)
{
  Point pt_dest = unit.m_location;
  if (invalid_point(pt_dest)) ERR("invalid point");
  
  switch (Rand.rand_uint4() % 3) {
  case 0: pt_dest.x--;
    break;
  case 1:
    break;
  case 2: pt_dest.x++;
    break;
  
  default:
    ERR("!");
    break;
  }

  switch (Rand.rand_uint4() % 3) {
  case 0: pt_dest.y--;
    break;
  case 1:
    break;
  case 2: pt_dest.y++;
    break;
  
  default:
    ERR("!");
    break;
  }

  if (area_empty(pt_dest.x, pt_dest.y)) {
	move(unit.m_obj_id, pt_dest);
  }
}

AI_Object* AI_Board::find_obj(sint4 obj_id)
{
  object_map::iterator iter = m_objs[FRIENDLY_SIDE].find(obj_id);
  if (iter != m_objs[FRIENDLY_SIDE].end())
	return &(iter->second);
  
  iter = m_objs[ENEMY_SIDE].find(obj_id);
  if (iter != m_objs[ENEMY_SIDE].end())
	return &(iter->second);
  
  return NULL;  
}

const AI_Object* AI_Board::find_obj(sint4 obj_id) const
{
  object_map::const_iterator iter = m_objs[FRIENDLY_SIDE].find(obj_id);
  if (iter != m_objs[FRIENDLY_SIDE].end())
	return &(iter->second);
  
  iter = m_objs[ENEMY_SIDE].find(obj_id);
  if (iter != m_objs[ENEMY_SIDE].end())
	return &(iter->second);
  
  return NULL;
}

void AI_Board::get_random_location_in_fog(Point &pt) const
{
  // todo: prefer 'likely' locations
  
  // for now, just pick a random spot in the fog
  // however, prefer unexplored fog...
  
  // todo: a less random approach (would be faster)
  const uint4 max_tries = 100;
  
  const uint4 unexplored_only_tries = 10;
  
  FORU (i, max_tries) {
	pt.x = Rand.rand_uint4() % m_board_width;
	pt.y = Rand.rand_uint4() % m_board_height;
	
	if (i + 1 >= max_tries) {
	  //cout << "warning: too many tries getting random location" << endl;
	  //pt.x = -1;
	  //pt.y = -1;
      
      // just go to any random location...
      
	  return;
	}
	
	const AI_Board_Area &area = get_c(pt);
	
	if (area.is_obstacle())
	  continue;
	
	if ((i < unexplored_only_tries) && (area.is_explored()))
	  continue;  // skip this explored area until <i> is greater

	if (!area.is_visible()) {
	  if (!area.is_valid()) 
		ERR("!");  // unexpected
		
	  // point found
	  break;
	}
  }
}

bool AI_Board::get_nearby_unexplored_point(Point &pt, const Point &pt_start) const
{
  // use random (inefficient...) algorithm for now
  const uint4 max_tries = 10;
  
  const sint4 dist = 10;
  
  FORU (i, max_tries) {
	pt.x = pt_start.x - dist + (Rand.rand_uint4() % (2 * dist));
	pt.y = pt_start.y - dist + (Rand.rand_uint4() % (2 * dist));
	
	if (i + 1 >= max_tries) {
      // fail: todo - try not to fail
	  return false;
	}
	
	const AI_Board_Area &area = get_c(pt);
	if (!area.is_valid())
      continue;
    
	if (area.is_obstacle())
	  continue;
	
	if (area.is_explored())
	  continue;

	if (!area.is_visible()) {		
	  // point found
	  break;
	}
  }
  
  return true;
}

bool AI_Board::get_nearby_empty_point(Point &pt, const Point &pt_start) const
{
  // use random (inefficient...) algorithm for now
  const uint4 max_tries = 10;
  
  const sint4 dist = 3;
  
  FORU (i, max_tries) {
	pt.x = pt_start.x - dist + (Rand.rand_uint4() % (2 * dist));
	pt.y = pt_start.y - dist + (Rand.rand_uint4() % (2 * dist));
	
	if (i + 1 >= max_tries) {
      // fail: todo - try not to fail
	  break;
	}
	
	const AI_Board_Area &area = get_c(pt);
	if (!area.is_valid())
      continue;
    
	if (area.is_obstacle())
	  continue;

    // point found
    return true;
  }
  
  return false;
}

bool AI_Board::is_game_won() const
{
   if (m_post_id[FRIENDLY_SIDE] < 0)
     return false;
   
   const AI_Object* p = find_obj(m_post_id[FRIENDLY_SIDE]);
   if (NULL == p)
     return false;
   
   return p->m_has_flag;
}

bool AI_Board::is_game_lost() const
{
   if (m_post_id[ENEMY_SIDE] < 0)
     return false;
   
   const AI_Object* p = find_obj(m_post_id[ENEMY_SIDE]);
   if (NULL == p)
     return false;
   
   return p->m_has_flag;
}

bool AI_Board::get_flag_location(Point &pt, sint4 side) const
{
  if (invalid_point(m_flag_loc[side]))
    return false;
  
  pt = m_flag_loc[side];
  
  return true;
}

bool AI_Board::get_post_location(Point &pt, sint4 side) const
{
  if (invalid_point(m_post_loc[side]))
    return false;
  
  pt = m_post_loc[side];
  
  return true;
}


// Local Variables:
// c-file-style:"gnu"
// c-basic-offset:2
// indent-tabs-mode:nil
// tab-width:4
// End:
