// $Id: AI_PathFind.C,v 1.1.1.1 2004/10/11 09:09:20 orts_mburo Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#include "Global.H"
#include "AI_PathFind.H"

// pathfind library
#include "pathfind.h"
using namespace PathFind;

#include "AI_Client.H"

#define SEARCH_NODE_LIMIT (10000)

/** <global> pathfinding function */
bool find_path(
  AI_Path &path,
  const AI_Board &board,
  Point pt_start,
  Point pt_goal)
{
  auto_ptr<Search> search;
  search.reset(new AStar<MarkerFastClear,
    AStarOpenBuckets<MarkerFastClear,AStarCompareF>,
    AStarClosedLookup<MarkerFastClear> >());
  
  search->setNodesLimit(SEARCH_NODE_LIMIT);
  AI_PathFind_Environment pfe;
  
  pfe.attach_board(&board);
  node_id_type start_id = pfe.point_to_id(pt_start);
  node_id_type goal_id = pfe.point_to_id(pt_goal);
  
  pfe.set_start(start_id);
  pfe.set_goal(goal_id);
  if (!search->findPath(pfe, start_id, goal_id))
    return false;
  
  path.set(search->getPath(), board.get_width());
  
  return true;
}


//---------------------------------------------------------------------
// AI_PathFind_Environment Implementation
//---------------------------------------------------------------------

AI_PathFind_Environment::AI_PathFind_Environment()
:
  PathFind::Environment(),

  m_board(NULL),
  m_start(-1),
  m_goal(-1)
{
  // ...
}

void AI_PathFind_Environment::attach_board(const AI_Board *board)
{
  if (NULL == board) ERR("null pointer - expected board object");
    
  m_board = board;
}

void AI_PathFind_Environment::detach_board()
{
  m_board = NULL;
}

node_id_type AI_PathFind_Environment::point_to_id(const Point &pt) const
{
  uint4 width = m_board->get_width();
  
  return (pt.y * width) + pt.x;
}

Point AI_PathFind_Environment::id_to_point(node_id_type id) const
{
  uint4 width = m_board->get_width();
  
  return Point(id % width, id / width);
}

void AI_PathFind_Environment::set_start(node_id_type id)
{
  m_start = id;
}

void AI_PathFind_Environment::set_goal(node_id_type id)
{
  m_goal = id;
}


//---------------------------------------------------------------------
// PathFind::Environment methods
//---------------------------------------------------------------------

int AI_PathFind_Environment::getHeuristic(int start, int target) const
{
  Point a = id_to_point(start);
  Point b = id_to_point(target);
  
  sint4 dx = abs(a.x - b.x);
  sint4 dy = abs(a.y - b.y);
  
  return dx + dy;
}

int AI_PathFind_Environment::getMaxCost() const
{
  return 1;
}

int AI_PathFind_Environment::getMinCost() const
{
  return 1;
}

int AI_PathFind_Environment::getNumberNodes() const
{
  return m_board->get_width() * m_board->get_height();
}
        
void AI_PathFind_Environment::getSuccessors(
  int nodeId, 
  int lastNodeId,
  vector<Successor>& result,
  bool usePruning /*= false*/) const
{
  const uint4 num_dirs = 4;
  
  result.reserve(num_dirs);
  result.clear();
  
  Point ptCurr = id_to_point(nodeId);
   
  for (uint4 i = 0; i < num_dirs; i++) {
    
    Point ptNext(ptCurr);
    
    switch (i) {
    case 0: ptNext.x--; break;
    case 1: ptNext.x++; break;
    case 2: ptNext.y--; break;
    case 3: ptNext.y++; break;
      
    default:
      ERR("unhandled case");
      break;
    }
    
    /*
    if (m_board->invalid_point(ptNext))
      continue;
    
    AI_Board_Area area = m_board->get_c(ptNext);
    if (!area.is_valid()) ERR("invalid area");
    
    if (area.is_obstacle())
      continue;
    */
    
    node_id_type id_next = point_to_id(ptNext);
    if (usePruning && (id_next == lastNodeId))
      continue;
    
    if ((id_next != m_start) && (id_next != m_goal) && (!m_board->area_empty(ptNext)))
      continue;
    
    result.push_back(Successor(id_next, 1));
  }
}
        
bool AI_PathFind_Environment::isValidNodeId(node_id_type nodeId) const
{
  Point pt = id_to_point(nodeId);
  return (m_board->valid_point(pt));
}


//---------------------------------------------------------------------
// Private methods
//---------------------------------------------------------------------

// ...


// Local Variables:
// c-file-style:"gnu"
// c-basic-offset:2
// indent-tabs-mode:nil
// tab-width:4
// End:
